/*
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.backup.plugins;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.SuspendableTask;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.kappich.annotations.NotNull;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.*;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

/**
 * TBD Dokumentation
 *
 * @author Kappich Systemberatung
 */
public abstract class AbstractFileSystemBackup implements BackupModul {
	public static final String PROP_BACKUP_DIR = "sicherungsVerzeichnis";
	protected static int MAX_CONTAINERS_PER_DIR = 1000;
	protected static long MAX_MEDIUM_SIZE_KB = 4300L * 1024L;
	protected Debug logger = Debug.getLogger();
	protected long currentSpaceOccupiedByZipfiles = 0;
	protected long currentSpaceOccupiedByContainerFiles = 0;
	protected String currentBackupPath;
	protected String backupBasePath = null;
	private int currentContIndex = 0;
	private List<File> currentContainersToBeSaved = new ArrayList<File>();

	public abstract void initialize(Properties props) throws Exception;

	public abstract boolean hasMediumCapacity(int mediumID, long containerSize, int mediumIndexSize) throws BackupException;

	public void backupContainer(int mediumID, long contID, File contFile) throws BackupException {

		try {
			// bis max-Grenze erreicht ist: File merken
			if(++currentContIndex % MAX_CONTAINERS_PER_DIR != 0) {
				currentContainersToBeSaved.add(contFile);
				currentSpaceOccupiedByContainerFiles += contFile.length();
			} // wenn max-Grenze erreicht ist: Zippen, Speichern und neues Verzeichnis bestimmen
			else {
				currentContainersToBeSaved.add(contFile);
				currentSpaceOccupiedByContainerFiles += contFile.length();
				backupContainerList();
			}
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
	}


	@Override
	public void openMediumBeforeBackup(int mediumID, String backupRunID) throws BackupException {
		String tmpMediumPath = backupBasePath + File.separator + "Medium_" + mediumID;
		if(!new File(tmpMediumPath).exists()) {
			new File(tmpMediumPath).mkdir();
		}
		currentBackupPath = tmpMediumPath + File.separator + backupRunID;
		if(!new File(currentBackupPath).exists()) {
			new File(currentBackupPath).mkdir();
		}
	}

	@Override
	@SuppressWarnings("unchecked")
	public void closeMediumAfterBackup(int mediumID, File indexFile) throws BackupException {
		try {
			if(currentContainersToBeSaved.size() != 0) {
				// die letzten Container noch zippen und schreiben:
				backupContainerList();
			}

			// indexFile abspeichern
			File newIndexFile = new File(currentBackupPath, indexFile.getName());
			FileInputStream is = new FileInputStream(indexFile);
			FileOutputStream os = new FileOutputStream(newIndexFile);
			copyStream(is, os);
			os.close();
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
		finally {
			currentContainersToBeSaved.clear();
			currentSpaceOccupiedByContainerFiles = 0;
			currentSpaceOccupiedByZipfiles = 0;
		}
	}

	public InputStream restoreContainer(int mediumID, String contFileName) throws BackupException {

		String backupPath = getBackupPath(mediumID);
		logger.fine("Suche nach Container " + contFileName + " im Verzeichnis: " + backupPath);

		long contID = ContainerFile.getContID(contFileName);
		File zipFile = null;
		try {
			List<File> backupRuns = Arrays.asList(Util.listDirectories(backupPath));
			Collections.sort(backupRuns);

			// indexFile auspacken und nachschauen, ob Container enthalten ist
			boolean containerFound = false;
			int i = backupRuns.size() - 1;
			String backupRun = "";
			while(!containerFound && i >= 0) {
				backupRun = backupRuns.get(i--).getName();

				File indexFile = new File(backupPath + File.separator + backupRun, BackupTask.INDEXFILE_NAME);
				if(!indexFile.exists()) {
					logger.info("kein Index-File gefunden: " + indexFile.getAbsolutePath());
					continue;
				}
				
				RandomAccessFile raf = new RandomAccessFile(indexFile, "r");
				for(String line; (line = raf.readLine()) != null;) {
					if(line.contains(contFileName)) {
						containerFound = true;
						break;
					}
				}
				raf.close();
			}
			if(!containerFound) {
				logger.error("Container " + contFileName + " konnte nicht auf MediumId=" + mediumID + " gefunden werden.");
				throw new BackupException("Container " + contFileName + " konnte nicht auf MediumId=" + mediumID + " gefunden werden.");
			}

			// nach dem richtigen Zip-File suchen:
			File[] zipfiles = new File(backupPath, backupRun).listFiles(
					new FileFilter() {

						public boolean accept(File pathname) {
							return pathname.getName().endsWith(".zip");
						}
					}
			);

			for(i = 0; i < zipfiles.length; i++) {
				String name = zipfiles[i].getName();
				String firstContainerNameAbb = name.substring(0, name.indexOf("-"));
				String lastContainerNameAbb = name.substring(name.indexOf("-") + 1, name.length() - 4);
				long firstContID = Long.parseLong(firstContainerNameAbb.substring(2, firstContainerNameAbb.length()));
				long lastContID = Long.parseLong(lastContainerNameAbb.substring(2, lastContainerNameAbb.length()));

				if(contID >= firstContID && contID <= lastContID) {
					zipFile = new File(backupPath + File.separator + backupRun, name);
					break;
				}
			}
			if(zipFile == null) {
				logger.warning("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.");
				throw new BackupException("Zip-Datei mit ContainerId=" + contID + " konnte nicht gefunden werden.");
			}
		}
		catch(Exception e) {
			logger.warning("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.", e.getMessage());
			throw new BackupException("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.");
		}
		try {
			return detachFileFromZip(zipFile, contFileName);
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
	}

	@NotNull
	public abstract String getBackupPath(int mediumID);

	public List<String> getAllContFileNames(SuspendableTask task, int mediumID, int maxLimit) throws BackupException {

		String backupPath = backupBasePath + File.separator + "Medium_" + mediumID;

		File tmpBackupPathFile = new File(backupPath);
		if(!tmpBackupPathFile.exists()) {
			logger.error("BackupPath " + backupPath + " existiert nicht.");
			throw new MediumNotAccessibleException(mediumID);
		}

		List<String> result = new ArrayList<String>();
		List<File> backupRuns = Arrays.asList(Util.listDirectories(backupPath));
		Collections.sort(backupRuns);

		int counter = 1;
		String backupRun = "";
		for(int i = backupRuns.size() - 1; i >= 0; i--) {
			task.suspendTaskIfNecessary();	//Task anhalten, falls Archivsystem überlastet
			backupRun = backupRuns.get(i).getName();

			File indexFile = new File(backupPath + File.separator + backupRun, BackupTask.INDEXFILE_NAME);
			if(!indexFile.exists()) {
				continue;
			}
			try {
				RandomAccessFile raf = new RandomAccessFile(indexFile.getAbsolutePath(), "r");
				for(String line; (line = raf.readLine()) != null;) {
					result.add(line.substring(line.lastIndexOf(File.separator) + 1));
					if(maxLimit != -1 && counter++ >= maxLimit) {
						return result;
					}
				}
				raf.close();
			}
			catch(Exception e) {
				Debug.getLogger().error("FileSystemBackup: Indexdatei " + indexFile.getAbsolutePath() + " konnte nicht gelesen werden.", e.getMessage());
			}
		}
		return result;
	}

	/**
	 * Entpackt eine ContainerDatei aus einem Zip-File.
	 *
	 * @param zipFile
	 * @param contFileName
	 *
	 * @return File
	 *
	 * @throws Exception
	 */
	private InputStream detachFileFromZip(File zipFile, String contFileName) throws Exception {
		try {
			ZipInputStream zin = new ZipInputStream(new BufferedInputStream(new FileInputStream(zipFile)));
			ZipEntry zen;
			while((zen = zin.getNextEntry()) != null) {
				if(zen.getName().equals(contFileName)) {
					return zin;
				}
			}
		}
		catch(IOException e) {
			throw new BackupException(e);
		}
		throw new BackupException(
				"Fehler beim Wiederherstellen: Datei '" + contFileName + "' konnte nicht aus Zip-Datei '" + zipFile.getAbsolutePath() + "' extrahiert werden."
		);
	}

	/**
	 * Zippt die in der Containerlist aufgelisteten Dateien und legt sie in den <code>currentBackupPath</code>.
	 *
	 * @throws Exception
	 */
	private void backupContainerList() throws Exception {

		File backupDir = new File(currentBackupPath);

		Collections.sort(currentContainersToBeSaved, new FileNameComparator());

		String nameFirst = currentContainersToBeSaved.get(0).getName();
		String nameLast = currentContainersToBeSaved.get(currentContainersToBeSaved.size() - 1).getName();
		String containerZipName = nameFirst.substring(0, nameFirst.indexOf("."));
		containerZipName += "-" + nameLast.substring(0, nameLast.indexOf("."));

		ZipOutputStream out;
		File zipFile = new File(backupDir, containerZipName + ".zip");
		out = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(zipFile)));
		for(File file : currentContainersToBeSaved) {
			ZipEntry entry = new ZipEntry(file.getName());
			entry.setSize(file.length());
			entry.setTime(file.lastModified());
			out.putNextEntry(entry);
			copyStream(new FileInputStream(file), out);
			out.closeEntry();
		}
		out.finish();
		out.close();
		currentSpaceOccupiedByZipfiles += zipFile.length();
		currentSpaceOccupiedByContainerFiles = 0;
		currentContainersToBeSaved.clear();
	}

	private void copyStream(InputStream in, OutputStream out) throws IOException {
		byte[] buffer = new byte[1024*8];
		int len;
		while((len = in.read(buffer)) >= 0) out.write(buffer, 0, len);

		in.close();
		// out.close();
	}

	public String toString() {
		return "FileSystem (" + backupBasePath + ")";
	}

	private class FileNameComparator implements Comparator<File> {

		public int compare(File f1, File f2) {
			return f1.getName().compareTo(f2.getName());
		}
	}
}
