/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2010 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.ContainerHdr;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.iter.DataIterator;
import de.bsvrz.ars.ars.persistence.iter.DeletedTreatment;
import de.bsvrz.ars.ars.persistence.iter.QueryDataSequence;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.IOException;
import java.util.HashSet;

/**
 * Task zur Bearbeitung von Archiv-Informationsanfragen.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveInfoQueryTask extends QueryTask {

	private static final boolean GAP = true, NO_GAP = false, NO_DIR_ACCESS = false;

	private static final int NO_MEDIUM_ID = -1;

	private ArchiveDataSpecification[] adsList;

	private byte[] indexMem;

	private int entryCounter;

	private DummyTask dummyTask = new DummyTask();    // zum autom., Verlaengern des Loeschschutzes

	public ArchiveInfoQueryTask(ArchiveManager archiveMgr, MultiTaskManager tMgr) {
		super(archiveMgr, tMgr);
	}

	private void analyze() throws Exception {
		createQueryData();
		adsList = parseArchiveDataSpec(deserializer);
	}


	/**
	 * Beantwortet die Archivinformationsanfragen. Die einzelnen Archivinformationsanfragen aus {@link #adsList} werden bearbeitet und beantwortet.
	 *
	 * @throws IOException
	 * @throws DataNotSubscribedException
	 * @throws SendSubscriptionNotConfirmed
	 */
	private void sendResponse() throws IOException, DataNotSubscribedException {
		indexMem = new byte[INDEX_MEMORY_SIZE];
		bosResult.reset();            // genuegt, man muss nicht jedesmal einen neuen Serialisierer anlegen

		serializer.writeByte(1);        // erfolgreich (nehmen wir erstmal an...)
		serializer.writeInt(0);            // Größe Liste (wird später besetzt)
		entryCounter = 0;

		DataIdentNode din = null;
		int adsCounter = 0;

		for(ArchiveDataSpecification ads : adsList) {
			suspendTaskIfNecessary();        //Task anhalten, falls Archiv überlastet.
			if(shouldTerminate()) break;

			long obj = ads.getObject().getId();
			long atg = ads.getDataDescription().getAttributeGroup().getId();
			long asp = ads.getDataDescription().getAspect().getId();
			short sv = ads.getDataDescription().getSimulationVariant();
			ArchiveTimeSpecification ats = ads.getTimeSpec();

			try {
				if(ats.isStartRelative()) throw new Exception("Relative Intervallgrenzen sind in Archiv-Informationsanfragen nicht erlaubt.");
				QueryDataSequence sequence = new QueryDataSequence(getArchMgr(), new IdDataIdentification(ads), ads.getDataKinds(), ats, ads.getSortOrder(), DeletedTreatment.MinMax, this, tempMem);

				DataIterator iterator = sequence.iterator();

				ContainerDataResult result = new ContainerDataResult();

				TimingType tt = ats.getTimingType();

				long curDIdx = -1, curATime = -1, curDTime = -1;                        // Werte des aktuellen Durchlaufs
				long lastDIdx = -1, lastDTime = -1, lastATime = -1, lastContID = -1;    // Werte des vorherigen Durchlaufs
				long lbDIdx = -1, lbATime = -1, lbDTime = -1;                            // Untere Grenze der Ausgabezeile
				int lastMedienID = -1, curMedienID = -1;                            // letzte und aktuelle (wg. Plattenzugriff) medienID
				boolean lastDirAccess = true, curDirAccess = true;                // letzter und aktueller (wg. Plattenzugriff) Direktzugriff

				HashSet<Long> alreadyLsExtended = new HashSet<>();
				// Hauptschleife:
				while(!iterator.isEmpty()) {
					if(shouldTerminate()) throw new Exception("Das Archivsystem wurde beendet.");

					iterator.peek(result);

					curDIdx = result.getDataIndex();
					curATime = result.getArchiveTime();
					curDTime = result.getDataTime();

					if(lbDIdx == -1) {
						// Untergrenzen initialisieren
						lbDIdx = curDIdx;
						lbATime = curATime;
						lbDTime = curDTime;
					}

					boolean blockAlreadyWritten = false;
					if(result.getContainerID() != lastContID) {
						// erster oder neuer Container
						lastMedienID = curMedienID;
						curMedienID = getMedienID(iterator);
						lastDirAccess = curDirAccess;
						curDirAccess = getDirAccess(iterator);

						if(lastContID != -1 && (lastMedienID != curMedienID || lastDirAccess != curDirAccess)) {
							// neuer Container (mindestens der 2.) mit anderen access-/medienID-Werten
							addEntry(lbDIdx, lbATime, lbDTime, lastDIdx, lastATime, lastDTime, tt, NO_GAP, lastDirAccess, lastMedienID, adsCounter);

							lbDIdx = curDIdx;
							lbATime = curATime;
							lbDTime = curDTime;
							blockAlreadyWritten = true;
						}
						lastContID = result.getContainerID();
					}

					if(isGap(result, lastDIdx)) {
						if(!blockAlreadyWritten) {
							// Zeile für Zustand bis zur Luecke (falls nicht schon wg. Containerwechsel ausgegeben):
							addEntry(lbDIdx, lbATime, lbDTime, lastDIdx, lastATime, lastDTime, tt, NO_GAP, curDirAccess, curMedienID, adsCounter);
						}

						// Zeile für die Luecke:
						addEntry(lastDIdx, lastATime, lastDTime, curDIdx, curATime, curDTime, tt, GAP, NO_DIR_ACCESS, NO_MEDIUM_ID, adsCounter);

						// naechste Zeile geht beim Lueckenende los:
						lbDIdx = curDIdx;
						lbATime = curATime;
						lbDTime = curDTime;
					}
					lastDIdx = curDIdx;
					lastATime = curATime;
					lastDTime = curDTime;

					// Archivinformationsanfrage loest automatische Loeschschutzverlaengerung aus (gemaess TAnfArS)
					if(alreadyLsExtended.add(result.getContainerID())) {
						extendLoeschschutz(obj, atg, asp, sv, result.getDataKind(), result.getContainerID());
					}
					iterator.remove();
				}

				// letzten Eintrag schreiben:
				addEntry(lbDIdx, lbATime, lbDTime, curDIdx, curATime, curDTime, tt, NO_GAP, curDirAccess, curMedienID, adsCounter);
				iterator.close();
			}
			catch(Exception e) {
				_debug.error("Fehler beim Bearbeiten der Archivinformationsanfrage", e);
				printError("Fehler beim Bearbeiten der Archivinformationsanfrage: " + e.getMessage());
				sendResultData(bosResult.toByteArray(), MessageType.QUERY_INFO_RESULT);
				return;
			}
			adsCounter++;
		}
		sendResultData(insertNumOfEntries(entryCounter), MessageType.QUERY_INFO_RESULT);
	}

	private boolean isGap(final ContainerDataResult result, final long lastDIdx) {
		return (lastDIdx != -1 && Util.dIdxNoModBits(result.getDataIndex()) - Util.dIdxNoModBits(lastDIdx) > 1) && !result.getDataKind().isDelayed();
	}

	private void extendLoeschschutz(long obj, long atg, long asp, int sv, ArchiveDataKind adk, long contID) throws PersistenceException {
		try {
			persMgr.accessContainer(dummyTask, obj, atg, asp, sv, adk, contID);
			if(persMgr.isContainerClosed(dummyTask)) {
				persMgr.setAutoExtendLoeschutz(dummyTask, RestoreTask.getLOESSCHUTZ_VERL(), t_lsMax);
				persMgr.writeContainerHeader(dummyTask);
			}
		}
		finally {
			persMgr.leaveContainer(dummyTask);
		}
	}

	private boolean getDirAccess(DataIterator iterator) throws PersistenceException {
		return !persMgr.getContainerHeaderParamAsBoolean(this, iterator.getCurContFileIdx(), ContainerHdr.CHP_DELETED);
	}

	private int getMedienID(DataIterator iterator) throws PersistenceException {
		String hdrMedID = persMgr.getContainerHeaderParamAsString(this, iterator.getCurContFileIdx(), ContainerHdr.CHP_MEDIUM_ID);
		if(!hdrMedID.equals(ContainerHdr.MEDIUM_UNSAVED) && !hdrMedID.equals(ContainerHdr.MEDIUM_DESTROYED)) {
			return Integer.parseInt(hdrMedID);
		}
		else {
			return -1;
		}
	}

	private long ttVal(TimingType tt, long dIdx, long aTime, long dTime) {
		return tt.equals(TimingType.DATA_TIME) ? dTime : (tt.equals(TimingType.ARCHIVE_TIME) ? aTime : dIdx);
	}

	private byte[] insertNumOfEntries(int numOfEntries) throws IOException {
		byte[] resultBytes = bosResult.toByteArray();
		bosResult.reset();
		serializer.writeInt(numOfEntries);
		System.arraycopy(bosResult.toByteArray(), 0, resultBytes, 1, 4);
		return resultBytes;
	}

	private void printError(String msg) throws IOException {
		bosResult.reset();
		serializer.writeByte(0);    // Fehler
		serializer.writeString(msg);
	}

	private void addEntry(
			long startDIdx, long startAT, long startDT, long endDIdx, long endAT, long endDT, TimingType tt, boolean gap, boolean direct, int volID, int idx)
			throws IOException {
		// System.out.println(startDIdx + "-" + endDIdx + " " + (gap ? "GAP":"DAT")+ " " + (direct ? "dac":"nodac"));
		serializer.writeLong(ttVal(tt, startDIdx, startAT, startDT));      // 3.
		serializer.writeLong(ttVal(tt, endDIdx, endAT, endDT));      // 4.
		serializer.writeByte(tt.equals(TimingType.DATA_TIME) ? 1 : (tt.equals(TimingType.ARCHIVE_TIME) ? 2 : 3));    // 5. TT
		serializer.writeByte(gap ? 1 : 0);        // 6. Gap
		serializer.writeByte(direct ? 1 : 0);    // 7. Access
		serializer.writeInt(volID);                // 8. volID
		serializer.writeInt(idx);                // 9.
		entryCounter++;
	}

	public void work() {
		//logger.finer(getName() + " Erhaelt Task: ", resultData);
		init();
		try {
			analyze();
			subscribeSender();
			// System.out.println("ArchiveInfoQueryTask.work");

			final SendingState state;
			// Nach 120 Sekunden ohne Sendesteuerung wird der Auftrag abgebrochen
			final long timeoutTime = System.currentTimeMillis() + (120 * 1000);
			synchronized(syncObj) {
				while(sendingState == SendingState.STOP) {
					final long delta = timeoutTime - System.currentTimeMillis();
					if(delta <= 0) break;
					syncObj.wait(delta);
				}
				state = sendingState;
			}

			if(state == SendingState.SEND) {
				sendResponse();
			}
		}
		catch(InterruptedException e) {
			_debug.finer(getName() + " Task durch Interrupt beendet.");
		}
		catch(Exception e) {
			// resultData wird vor Aufruf der Debug-Methode in einen String konvertiert, um verschachtelte Debug-Aufrufe mit Deadlock-Gefahr zu vermeiden
			_debug.error(getName() + " Fehler beim Bearbeiten der Archiv-Informationsanfrage: " + e.getMessage() + ", Anfrage: " + resultData.toString());
		}
		finally {
			unsubscribeSender(); // Als Sender der Archiv-Antwort abmelden
			persMgr.removeContainerFileObjects(this, MIN_CONTAINER_FILES); // ContainerFiles aufraeumen
			adsList = null;
			indexMem = null;
			bosResult.reset();
		}
	}


}
