/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.sys.funclib.losb.ringbuffer.ArchiveRingBuffer;
import de.bsvrz.sys.funclib.losb.util.Util;

/**
 * Task zur Archivierung nachgefordeter Daten. Unterscheidet sich hauptsaechlich in der Behandlung von Datenluecken von {@link ArchiveTask}.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ArchiveRequestedTask extends ArchiveTask {

	private ArchiveData ad;

	/**
	 * Erzeugt einen ArchiveRequestedTask
	 *
	 * @param archiveMgr Archiv-Manager
	 * @param rodl       Ring-of-Death-Listener
	 * @param tskIndex   Index des Tasks
	 * @param inQueue    Eingangs-Warteschlange
	 */
	public ArchiveRequestedTask(ArchiveManager archiveMgr,int tskIndex, ArchiveRingBuffer inQueue) {
		super(archiveMgr, tskIndex, inQueue);
	}

	/**
	 * Alle Instanzvariablen vorbelegen.
	 *
	 * @param archiveData Archivdatensatz
	 * @param archTime    Archivzeitstempel
	 *
	 * @throws DIdNodeNotFoundException
	 */
	private void presetData(ArchiveData archiveData, long archTime) throws DIdNodeNotFoundException {
		ad = archiveData;
		DataDescription dd = ad.getDataDescription();
		objID = ad.getObject().getId();
		atg = dd.getAttributeGroup();
		aspID = dd.getAspect().getId();
		sv = 0;
		arsTime = archTime;
		dataTime = ad.getDataTime();
		dataIdx = ad.getDataIndex();
		adk = ad.getDataKind().isDelayed() ? NN : NA;
		didPath = persMgr.getPathName(objID, atg.getId(), aspID, 0, adk);
		din = didTree.getDataIdentNode(objID, atg.getId(), aspID, 0);
	}

	public void work(Object archiveData, long archTime) {
		logReceivedData((ArchiveData)archiveData);

		// Synchronisierung ueber den DataIdentNode:
		boolean criticalAreaEntered = false;
		try {
			// Flags, IDs, Variable besetzen, DataIdentNode holen, etc.:
			presetData((ArchiveData)archiveData, archTime);

			criticalAreaEntered = din.enterCriticalArea();
			if(!criticalAreaEntered) {
				_debug.warning("Archivierung nachgeforderter Daten abgebrochen: " + ad);
				return;
			}

			assignIndexes(REQUESTED);			// Indexe zuweisen, falls nicht schon geschehen
			lookupLastDataIdxes(REQUESTED);		// Letzte Datenindexe, Datenzeiten nachsehen

			openContID = din.getOpenContID(adk);
			dataEntryLen = persMgr.serializeReqData(this, ad, arsTime, compresser);
			(storeData() ? _successCount : _failCount).incrementAndGet();
		}
		catch(DIdNodeNotFoundException | IndexException | PersistenceException e) {
			handleErr(e, Util.ad2Str(ad));
		}
		catch(Throwable e) {
			handleSevereErr(e, Util.ad2Str(ad));
		}

		finally {
			if(din != null && criticalAreaEntered) din.exitCriticalArea();
		}
	}

	protected void logReceivedData(ArchiveData ad) {
		_debug.finer(getName() + ": nachgeforderten Datensatz zur Archivierung erhalten: ", Util.ad2Str(ad));
	}

}
