/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2010 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.mgmt.tasks;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.archive.*;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.dataSerializer.NoSuchVersionException;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.losb.datk.MessageType;
import de.bsvrz.sys.funclib.losb.datk.PidQuery;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.util.ByteIO;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.BlockingQueue;

/**
 * Diese Klasse ist die Basis-Klasse aller Single-Tasks, die mit dem DAV kommunizieren.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision$ / $Date$ / ($Author$)
 */
public abstract class SingleClientSenderTask extends SingleTask implements ClientSenderInterface {

	/** Zeit in Millisekunden, die auf eine positive Sendesteuerung gewartet wird */
	public static final int SENDCONTROL_TIMEOUT = 30000;

	/** Aktueller Zustand der Sendesteuerung. */
	private boolean sendingAllowed = false;

	/** Spezielles Objekt zur Synchronisiserung. Immer dieses anstatt this verwenden! */
	protected Object syncObj = new Object();

	/** Serialisiererversion der anfragenden Applikation, mit der auch die Antwort erstellt werden muss. */
	protected int serializerVersion;

	/** Attributgruppe der Archiv-Anfrage und -Antwort. */
	protected AttributeGroup atgQuery;

	/** DataDescription der Archiv-Antwort. */
	protected DataDescription ddResponse;

	/** Applikations-Objekt des Absenders der Archiv-Anfrage. */
	protected SystemObject queryAppObj;

	/** Anfrage-Index der anfragenden Applikation zur Archiv-Anfrage. */
	protected int queryIdx;

	private ByteArrayOutputStream resultStream = new ByteArrayOutputStream();

	private Serializer serializer;

	protected Deserializer deserializer;

	/**
	 * Erzeugt einen neuen SingleClientSenderTask.
	 * @param archiveMgr Der Archiv-Manager.
	 * @param inputData  Eingangswarteschlange
	 *
 	 */
	public SingleClientSenderTask(ArchiveManager archiveMgr, BlockingQueue<Object> inputData) {
		super(archiveMgr, inputData);
	}

	@Override
	public void getObjectsFromDav() {
		atgQuery = getArchMgr().getDataModel().getAttributeGroup(PidQuery.ATG_QUERY_PID);
		ddResponse = new DataDescription(atgQuery, getArchMgr().getDataModel().getAspect(PidQuery.ASP_RESPONSE_PID));
	}

	/**
	 * Initialisiert den Task für eine neue Archivanfrage.
	 *
	 * @throws NoSuchVersionException
	 */
	protected void init(Data data) throws NoSuchVersionException {
		queryAppObj = data.getReferenceValue(PidQuery.ATT_SENDER_NAME).getSystemObject();
		queryIdx = data.getUnscaledValue(PidQuery.ATT_QUERY_IDX_NAME).intValue();

		byte[] dataArray = data.getUnscaledArray(PidQuery.ATT_DATA_NAME).getByteArray();
		serializerVersion = (int)ByteIO.readSignedInt4Bytes(dataArray, 0); // Erste 4 Bytes = Serialisierer-Version
		InputStream bis = new ByteArrayInputStream(dataArray, 4, dataArray.length - 4);

		// Beide Teile bereits hier erzeugen, damit eine falsche Version zu Beginn bemerkt wird
		deserializer = SerializingFactory.createDeserializer(serializerVersion, bis);
		serializer = SerializingFactory.createSerializer(serializerVersion, resultStream);
	}

	protected void sendErrorResponse(String errorMessage) {
		sendResponse(false, errorMessage);
	}

	protected void sendSuccessResponse() {
		sendResponse(true, "");
	}

	protected void sendErrorResponse(int messageTypeResult, String errorMessage) {
		sendResponse(messageTypeResult, false, errorMessage);
	}

	protected void sendSuccessResponse(int messageTypeResult) {
		sendResponse(messageTypeResult, true, "");
	}

	private void sendResponse(boolean success, String msg) {
		sendResponse(getMessageTypeResult(), success, msg);
	}

	private void sendResponse(int messageTypeResult, boolean success, String msg) {
		try {
			sendingAllowed = false;
			subscribeSender();
			long start = System.currentTimeMillis();
			synchronized(syncObj) {
				while(!sendingAllowed) {
					syncObj.wait(SENDCONTROL_TIMEOUT);

					if(!sendingAllowed && (System.currentTimeMillis() - start >= SENDCONTROL_TIMEOUT)) {
						// Antwort kann nicht gesendet werden, weil keine Sendesteuerung innerhalb der Timeoutzeit empfangen wurde
						_debug.warning(
								getName() + ": Keine positive Sendesteuerung zum Verschicken der Antwort nach " + (SENDCONTROL_TIMEOUT / 1000)
								+ " Sekunden erhalten."
						);
						return;
					}
				}
			}
			Data data = archMgr.getDavCon().createData(atgQuery);
			data.getReferenceValue(PidQuery.ATT_SENDER_NAME).setSystemObject(archMgr.getConfigAuth());
			data.getUnscaledValue(PidQuery.ATT_QUERY_IDX_NAME).set(queryIdx);
			data.getUnscaledValue(PidQuery.ATT_MESSAGE_TYP_NAME).set(messageTypeResult);

			resultStream.reset();
			serializer.writeByte(success ? 1 : 0);
			if(!msg.equals("")) serializer.writeString(msg);
			else if(success){
				insertData(serializer);
			}

			byte[] message = resultStream.toByteArray();
			Data.NumberArray dataBytes = data.getUnscaledArray(PidQuery.ATT_DATA_NAME);
			dataBytes.set(message);

			ResultData rd = new ResultData(queryAppObj, ddResponse, System.currentTimeMillis(), data);
			archMgr.getDavCon().sendData(rd);
			resultStream.reset();
		}
		catch(InterruptedException e) {
			_debug.warning("Archivsystem beendet. Versenden der Antwort abgebrochen.");
		}
		catch(Exception e) {
			_debug.warning("Versenden der Antwort fehlgeschlagen.", e);
		}
		finally {
			unsubscribeSender();
		}
	}

	protected void insertData(final Serializer serializer) throws IOException {
		// Hier kann in überschriebenen Klassen etwas in den Serializer geschrieben werden um bei einer positiven Antwort noch Daten mitzugeben
	}

	private int getMessageTypeResult() {
		Class task = this.getClass();
		if(task == DeleteSimVarTask.class) {
			return MessageType.DELETE_RESULT;
		}
		else if(task == HeaderRestoreTask.class) {
			return MessageType.HEADER_RESTORE_RESULT;
		}
		else if(task == LZVTask.class) {
			return MessageType.DELETION_TIME_RESULT;
		}
		else if(task == RestoreTask.class) {
			return MessageType.RESTORE_RESULT;
		}
		else if(task == BackupTask.class) {
			return MessageType.BACKUP_RESULT;
		}
		else if(task == RequestGapTask.class) {
			return MessageType.REQUEST_DID_RESULT;
		}
		else if(task == QueryInfoTask.class) {
			return MessageType.NUM_QUERIES_INFO_RESULT;
		}
		return -1;
	}


	/**
	 * Meldet den Task als Sender für Archiv-Antworten an das angegebene Empfaenger-Applikations-Objekt an. Registriert den StreamMultiplexer für den Empfang von
	 * Flusskontroll-Steuerungspaketen.
	 *
	 * @throws OneSubscriptionPerSendData
 	 */
	protected void subscribeSender() throws OneSubscriptionPerSendData {
		ConnectionManager.subscribeSender(archMgr.getDavCon(), this, queryAppObj, ddResponse, SenderRole.sender());
	}

	/** Meldet den Task als Sender für Archiv-Antworten an das angegebene Empfaenger-Applikations-Objekt ab. */
	protected void unsubscribeSender() {
		try {
			ConnectionManager.unsubscribeSender(archMgr.getDavCon(), this, queryAppObj, ddResponse);
		}
		catch(Exception e) {
			_debug.warning(getName() + " Sende-Abmeldung fehlgeschlagen.", e);
		}
	}

	// ////////////////////////////////////////////////////////////////////////////////////////////////////////
	// ClientSenderInterface Methoden
	// ////////////////////////////////////////////////////////////////////////////////////////////////////////
	/** @see ClientSenderInterface#dataRequest(SystemObject,DataDescription,byte) */
	public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
		synchronized(syncObj) {
			sendingAllowed = (state == ClientSenderInterface.START_SENDING);
			syncObj.notifyAll();
		}
	}

	/** @see ClientSenderInterface#isRequestSupported(SystemObject,DataDescription) */
	public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
		return true; // Sendesteuerung ist erwuenscht und wird in dataRequest() verarbeitet.
	}

	protected ArchiveDataSpecification[] parseArchiveDataSpec(Deserializer ds, ResultData resultData) throws Exception {

		// ====== Analyse der empfangenen Daten erfolgt hier gemaess ======
		// ====== Datenkatalog Kapitel 7.2.3.12 ArchivAnfrageSchnittstelle ======

		int anzQueries = deserializer.readInt(); // Größe der Liste mit den ArchiveDataSpecification
		ArchiveDataSpecification[] result = new ArchiveDataSpecification[anzQueries];

		for(int i = 0; i < anzQueries; i++) {
			TimingType tt = Util.getTimingType(ds.readByte()); // Typ der Timingangabe

			boolean relative = ds.readByte() == 1; // Intervallstart relativ
			long start = ds.readLong(); // Intervallstart
			long end = ds.readLong(); // Intervallende

			ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(tt, relative, start, end);

			boolean oa = ds.readByte() == 1; // Online aktuell
			boolean on = ds.readByte() == 1; // Online nachgeliefert
			boolean na = ds.readByte() == 1; // Nachgefordert aktuell
			boolean nn = ds.readByte() == 1; // Nachgefordert nachgeliefert

			ArchiveDataKindCombination adkComb = Util.getADKCombination(oa, on, na, nn);
			if(adkComb == null) throw new Exception("Keine Datenart spezifiziert: oa=on=na=nn=0");

			int orderCode = ds.readInt();
			ArchiveOrder order = orderCode != 0 ? ArchiveOrder.getInstance(orderCode) : null;

			int reqOptCode = ds.readInt();
			ArchiveRequestOption reqOpt = reqOptCode != 0 ? ArchiveRequestOption.getInstance(reqOptCode) : null;

			AttributeGroup atg = (AttributeGroup)ds.readObjectReference(archMgr.getDataModel());
			Aspect asp = (Aspect)ds.readObjectReference(archMgr.getDataModel());
			short sv = ds.readShort();
			SystemObject obj = ds.readObjectReference(archMgr.getDataModel());

			result[i] = new ArchiveDataSpecification(timeSpec, adkComb, order, reqOpt, new DataDescription(atg, asp, sv), obj);
		}
		return result;
	}
}
