/*
 *
 * Copyright 2010 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence;

import de.bsvrz.sys.funclib.debug.Debug;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

/**
 * Ein ContainerBackupIndex ist eine Klasse, die innerhalb einer Datenindentifikationsverzeichnisses speichert, welche Container in diesem Verzeichnis noch zu
 * sichern sind, damit nicht jeder einzelne ContainerHeader geprüft werden muss.<br> Es wird eine Datei namens _backupindex.idx gespeichert, welche die
 * eigentlichen Einträge enthält. Diese besteht aus einem Header-Bereich un einem Datenbereich, wobei der Datenbereich unterteilt ist, in sortierte Daten und
 * unsortierte Daten. Zur Optimierung des Zugriffs, wird, falls es notwendig wird, der komplette Datenbereich sortiert.<br> Als Daten enthält ein
 * ContainerBackupIndex ContainerIDs mit einem zugehörigen Byte-Wert, der den Status dieses Container-Eintrags angibt (sichern oder ungültig). Sichern bedeutet,
 * dass der Container gesichert wird. Ungültig bedeutet, dass ein Eintrag ungültig ist. Ein Eintrag wird typischerweise als ungültig markiert, wenn der
 * entsprechende Container gesichert wurde, um diesen nicht erneut zu sichern. Einen Eintrag als ungültig zu markieren ist performancemäßig sinnvoller, als
 * ständig den gesamten Datenbereich neu zu schreiben um eventuelle Lücken zu schließen.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public final class ContainerBackupIndex {

	/** Maximale MedienID unbekannt */
	public static final int MEDIUM_ID_UNKNOWN = -1;

	/** Datenidentifikation wurde noch nie gesichert */
	public static final int MEDIUM_ID_UNSAVED = -2;

	/** Dateiname unter dem der Index gespchert wird */
	public static final String BACKUP_INDEX_FILE_NAME = "_backupindex.idx";

	private static final Debug _debug = Debug.getLogger();

	private final ContainerBackupIndexImplementation _impl;

	private final Object _implLock = new Object();

	private boolean _isOpened = true;

	/**
	 * Erstellt mit der angebenden ContainerBackupIndexImplementation einen neuen ContainerBackupIndex.
	 *
	 * @param implementation Implementation
	 */
	private ContainerBackupIndex(ContainerBackupIndexImplementation implementation) {
		_impl = implementation;
	}

	/**
	 * Gibt einen ContainerBackupIndex aus einem angegebenen Verzeichnis zurück
	 *
	 * @param path Verzeichnis
	 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
	 */
	public static ContainerBackupIndex getContainerBackupIndex(File path) {
		final ContainerBackupIndexImplementation implementation = ContainerBackupIndexImplementation.getContainerBackupIndex(path);
		if(implementation == null) return null;
		return new ContainerBackupIndex(implementation);
	}

	/**
	 * Gibt einen ContainerBackupIndex aus einem angegebenen Verzeichnis zurück
	 *
	 * @param path Verzeichnis
	 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
	 * @throws IOException IO-Fehler
	 */
	public static ContainerBackupIndex getContainerBackupIndex(String path) throws IOException {
		final ContainerBackupIndexImplementation implementation = ContainerBackupIndexImplementation.getContainerBackupIndex(path);
		if(implementation == null) return null;
		return new ContainerBackupIndex(implementation);
	}

	/**
	 * Erstellt einen neuen, leeren ContainerBackupIndex im angegebenen Verzeichnis. Ein eventuell bereits vorhandener wird überschrieben.
	 *
	 * @param path Verzeichnis
	 * @return der neue ContainerBackupIndex
	 * @throws IOException IO-Fehler
	 */
	public static ContainerBackupIndex createNewContainerBackupIndex(File path) throws IOException {
		return new ContainerBackupIndex(ContainerBackupIndexImplementation.createNewContainerBackupIndex(path));
	}

	/**
	 * Löscht den ContainerBackupIndex im angebenden Verzeichnis
	 *
	 * @param path Verzeichnis
	 * @return true wenn das Löschen erfolgreich war und der Index vorher existierte
	 */
	public static boolean deleteContainerBackupIndex(File path) {
		return ContainerBackupIndexImplementation.deleteContainerBackupIndex(path);
	}

	/**
	 * Prüft ob im angebenden Verzeichnis ein ContainerBackupIndex existiert
	 *
	 * @param path Verzeichnis
	 * @return true wenn er existiert
	 */
	public static boolean existsIn(File path) {
		return ContainerBackupIndexImplementation.getContainerBackupIndex(path) != null;
	}

	/**
	 * Prüft ob im angebenden Verzeichnis ein ContainerBackupIndex existiert
	 *
	 * @param path Verzeichnis
	 * @return true wenn er existiert
	 */
	public static boolean existsIn(String path) {
		return ContainerBackupIndexImplementation.getContainerBackupIndex(path) != null;
	}

	/**
	 * Gibt alle zu sichernden ContainerIDs im Index zurück
	 *
	 * @return Liste mit ContainerIDs
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public List<Long> getContainerIDs() throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.getContainerIDs();
		}
	}

	/**
	 * Entfernt eine ContainerID aus dem Index
	 *
	 * @param containerID ContainerID die zu entfernen ist.
	 * @return True falls erfolgreich, false falls nicht gefunden
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public boolean removeContainerID(long containerID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.removeContainerID(containerID, MEDIUM_ID_UNKNOWN);
		}
	}

	/**
	 * Entfernt eine ContainerID aus dem Index
	 *
	 * @param containerID ContainerID die zu entfernen ist.
	 * @param mediumID    Beim sichern eines Containers kann eine MediumID mit angegeben werden, die dann im Index vermerkt wird. Wird nur berücksichtigt, wenn
	 *                    es sich um eine positive Zahl handelt und nicht bereits eine größere MedienID in der Datei vorhanden ist
	 * @return True falls erfolgreich, false falls nicht gefunden
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public boolean removeContainerID(long containerID, final int mediumID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.removeContainerID(containerID, mediumID);
		}
	}

	/**
	 * Gibt zurück, ob eine Container-ID im Index enthalten ist
	 *
	 * @param containerID ContainerID die zu entfernen ist.
	 * @return True falls enthalten, false sonst
	 * @throws IOException IO-Fehler
	 */
	public boolean containsContainerID(long containerID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.containsContainerID(containerID);
		}
	}

	/**
	 * Fügt dem Index eine ContainerID hinzu
	 *
	 * @param containerID ContainerID die einzufügen ist
	 * @throws IOException           IO-Fehler
	 * @throws IllegalStateException Falls Index bereits geschlossen wurde
	 */
	public void appendContainerID(long containerID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			_impl.appendContainerID(containerID);
		}
	}

	/**
	 * Gibt die in der ContainerBackupIndex-Datei maximal gespeicherte MedienID zurück.
	 *
	 * @return MedienID
	 * @throws IOException IO-Fehler
	 */
	public int getMaxMedienID() throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			return _impl.getMaxMedienID();
		}
	}

	/**
	 * Setzt die MedienID auf einen festgelegten Wert (erlaubt auch das Zurücksetzen oder das Setzen von {@link ContainerBackupIndex#MEDIUM_ID_UNSAVED} bzw.
	 * {@link ContainerBackupIndex#MEDIUM_ID_UNKNOWN}
	 *
	 * @param mediumID MedienID
	 * @throws IOException IO-Fehler
	 */
	public void setMaxMedienID(int mediumID) throws IOException {
		synchronized(_implLock) {
			if(!_isOpened) throw new IllegalStateException("Dieser Index wurde bereits geschlossen");
			_impl.setMaxMedienID(mediumID);
		}
	}

	/** Schließt diesen ContainerBackupIndex. Muss immer nach Benutzung aufgerufen werden um kein mögliches Speicherleck zu verursachen */
	public void close() {
		synchronized(_implLock) {
			if(_isOpened) {
				_impl.removeReference();
				_isOpened = false;
			}
		}
	}

	@Override
	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}

	@Override
	public String toString() {
		synchronized(_implLock) {
			return "ContainerBackupIndex{" + _impl + ", open=" + _isOpened + '}';
		}
	}

	/**
	 * Eigentliche Implementierung des ContainerBackupIndex. wird von der {@link ContainerBackupIndex}-Klasse gekapselt um einen sicheren Referenzzähler zu
	 * ermöglichen. Zu den Zwecken der Threadsicherheit wird für jede ContainerBackupIndex-Datei maximal eine Instanz erzeugt und verwaltet. Jede Instanz kann
	 * von mehreren ContainerBackupIndex-Objekten verwendet werden. Zu diesem Zweck implementiert diese Klasse einen Referenzzähler.
	 */
	private static final class ContainerBackupIndexImplementation {

		private static final int CURRENT_FILE_VERSION = 1;
		/** Legt fest, dass der Container im Index geschlossen ist und gesichert werden kann (z.B. wenn die Parametrierung es verlangt) */
		private static final int CONTAINER_TO_SAVE = 0;
		/**
		 * Markiert einen ungültigen Index-Eintrag. Wenn ein Container aus dem Index gelöscht wird, wird er zunächst mit diesem Flag markiert Erst wenn die
		 * Anzahl der ungültigen Einträge einen bestimmten Wert überschreitet, wird die gesamte Datei neu geschrieben/komprimiert.
		 */
		private static final int CONTAINER_DELETED_FROM_INDEX = -1;
		/**
		 * Aus Synchronitätsgründen muss sichergestellt sein, dass wenn eine Indexdatei an zwei verschiedenen Stellen gleichzeitig angefragt wird, beide Threads
		 * das selbe Objekt erhalten. Daher werden hier die offenen Indexdateien verwaltet.
		 */
		private static final HashMap<File, ContainerBackupIndexImplementation> _openFiles = new HashMap<File, ContainerBackupIndexImplementation>();
		private static final int SIZE_INTEGER = 4;
		private static final int SIZE_BYTE = 1;
		private static final int SIZE_LONG = 8;
		private static final int HEADER_VERSION_SIZE = SIZE_INTEGER;
		private static final int HEADER_MEDIUM_ID_SIZE = SIZE_INTEGER;
		private static final int HEADER_NUM_SORTED_ENTRIES_SIZE = SIZE_INTEGER;
		private static final int HEADER_NUM_GAPS_SIZE = SIZE_INTEGER;
		private static final int HEADER_POS_MEDIUM_ID = HEADER_VERSION_SIZE;
		private static final int HEADER_POS_NUM_SORTED_ENTRIES = HEADER_POS_MEDIUM_ID + HEADER_MEDIUM_ID_SIZE;
		private static final int HEADER_POS_NUM_GAPS = HEADER_POS_NUM_SORTED_ENTRIES + HEADER_NUM_SORTED_ENTRIES_SIZE;
		private static final int HEADER_SIZE = HEADER_VERSION_SIZE + HEADER_MEDIUM_ID_SIZE + HEADER_NUM_SORTED_ENTRIES_SIZE + HEADER_NUM_GAPS_SIZE;
		private static final int ENTRY_SIZE = SIZE_LONG + SIZE_BYTE;
		private final Object _fileLock = new Object();
		private File _file;
		/** Anzahl der Referenzen */
		private int _referenceCount;

		private ContainerBackupIndexImplementation(final File file) {
			_file = file;
			_referenceCount = 1;
			synchronized(_openFiles) {
				_openFiles.put(_file, this);
			}
		}

		/**
		 * Gibt eine ContainerBackupIndexImplementation aus einem angegebenen Verzeichnis zurück
		 *
		 * @param path Verzeichnis
		 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
		 */
		public static ContainerBackupIndexImplementation getContainerBackupIndex(final File path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			if(file.exists()) {
				return getInstance(file);
			}
			return null;
		}

		/**
		 * Gibt eine ContainerBackupIndexImplementation aus einem angegebenen Verzeichnis zurück
		 *
		 * @param path Verzeichnis
		 * @return ein ContainerBackupIndex oder null falls er noch nicht existiert
		 */
		public static ContainerBackupIndexImplementation getContainerBackupIndex(final String path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			if(file.exists()) {
				return getInstance(file);
			}
			return null;
		}

		/**
		 * Holt zu einer angegebenen Datei eine ContainerBackupIndexImplementation-Instanz. Entweder aus der _openFiles-HashMap oder es wird ein neues Objekt
		 * erstellt
		 *
		 * @param file Datei
		 * @return die ContainerBackupIndexImplementation
		 */
		private static ContainerBackupIndexImplementation getInstance(File file) {
			synchronized(_openFiles) {
				final ContainerBackupIndexImplementation instance = _openFiles.get(file);
				if(instance != null) {
					instance.addReference();
					return instance;
				}
				return new ContainerBackupIndexImplementation(file);
			}
		}

		/**
		 * Erstellt eine neuen, leeren ContainerBackupIndexImplementation im angegebenen Verzeichnis. Ein eventuell bereits vorhandener wird überschrieben.
		 *
		 * @param path Verzeichnis
		 * @return der neue ContainerBackupIndex
		 * @throws IOException IO-Fehler
		 */
		public static ContainerBackupIndexImplementation createNewContainerBackupIndex(final File path) throws IOException {
			synchronized(_openFiles) {
				final File file = new File(path, BACKUP_INDEX_FILE_NAME);
				ContainerBackupIndexImplementation containerBackupIndex = getContainerBackupIndex(file);
				if(containerBackupIndex == null) containerBackupIndex = new ContainerBackupIndexImplementation(file);
				containerBackupIndex.reset();
				return containerBackupIndex;
			}
		}

		/**
		 * Löscht einen ContainerBackupIndex aus dem angegeben Verzeichnis
		 *
		 * @param path Verzeichnis
		 * @return true wenn das Löschen erfolgreich war und die Datei vorher existierte
		 */
		public static boolean deleteContainerBackupIndex(final File path) {
			final File file = new File(path, BACKUP_INDEX_FILE_NAME);
			return file.delete();
		}

		private void addReference() {
			synchronized(_openFiles) {
				_referenceCount++;
			}
		}

		private void removeReference() {
			synchronized(_openFiles) {
				_referenceCount--;
				if(_referenceCount == 0) {
					_openFiles.remove(_file);
					_file = null;
				}
			}
		}

		/**
		 * Entfernt alle Daten aus der ContainerBackupIndexImplementation
		 *
		 * @throws IOException IO-Fehler
		 */
		private void reset() throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					randomAccessFile.setLength(0L);
					randomAccessFile.writeInt(CURRENT_FILE_VERSION);
					randomAccessFile.writeInt(MEDIUM_ID_UNKNOWN);
					randomAccessFile.writeInt(0);
					randomAccessFile.writeInt(0);
				}
			}
		}

		/**
		 * Fügt einen Container hinzu
		 *
		 * @param containerID Container-ID
		 * @throws IOException IO-Fehler
		 */
		public void appendContainerID(long containerID) throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
					final int numOfSortedEntries = randomAccessFile.readInt();

					boolean alreadyExisting = containsContainerID(containerID, randomAccessFile, numOfSortedEntries);

					if(alreadyExisting) {
						_debug.warning("Container-ID " + containerID + " wird in Backup-Index eingetragen, ist dort aber bereits vorhanden", new Throwable());
						return;
					}

					if(randomAccessFile.length() == HEADER_SIZE + numOfSortedEntries * ENTRY_SIZE) {
						// Optimierung:
						// Es gibt keine unsortierten Einträge in der Datei.
						// Falls der einzufügende Eintrag größer ist, als der letzte Eintrag, bleibt die komplette Containerliste sortiert und man kann
						// NUM_SORTED_ENTRIES um eins erhöhen.
						final long lastContainerID;
						if(numOfSortedEntries == 0) {
							// Wenn es keine Einträge gibt ist die Datei beim Einfügen des ersten immer sortiert
							lastContainerID = -1;
						}
						else {
							// Wenn es bereits Einträge gibt muss der letzte untersucht werden
							randomAccessFile.seek(randomAccessFile.length() - ENTRY_SIZE);
							lastContainerID = randomAccessFile.readLong();
						}
						if(containerID > lastContainerID) {
							// Die neue ContainerID ist größer als die letzte, also NUM_SORTED_ENTRIES inkrementieren
							randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
							randomAccessFile.writeInt(numOfSortedEntries + 1);
						}
					}
					randomAccessFile.seek(randomAccessFile.length());
					randomAccessFile.writeLong(containerID);
					randomAccessFile.writeByte(CONTAINER_TO_SAVE);
				}
			}
		}

		/**
		 * Entfernt eine ContainerID aus dem Index
		 *
		 * @param containerID ContainerID die zu entfernen ist.
		 * @param mediumID    MediumID im Header, die evtl. erhöht werden muss
		 * @return True falls erfolgreich, false falls nicht gefunden
		 * @throws IOException IO-Fehler
		 */
		public boolean removeContainerID(long containerID, int mediumID) throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					final int fileMediumID = randomAccessFile.readInt();
					if(mediumID != MEDIUM_ID_UNSAVED && mediumID != MEDIUM_ID_UNKNOWN && mediumID > fileMediumID) {
						randomAccessFile.seek(HEADER_POS_MEDIUM_ID);
						randomAccessFile.writeInt(mediumID);
					}
					final int numOfSortedEntries = randomAccessFile.readInt();
					final int numOfGaps = randomAccessFile.readInt();
					final int numOfUnsortedEntries = (int) ((randomAccessFile.length() - HEADER_SIZE) / ENTRY_SIZE - numOfSortedEntries);

					if((numOfUnsortedEntries > 500 && numOfUnsortedEntries * 3 > numOfSortedEntries) || (numOfGaps > 4000
							&& numOfGaps * 3 > numOfSortedEntries * 2)) {
						// Wenn die Anzahl der deaktivierten Einträge (Lücken) 4000 überschreitet und die Anzahl der Lücken mindestens zwei Drittel
						// der Einträge ausmacht
						// sollte die Datei beim nächsten Zugriff "defragmentiert", also sortiert und komprimiert neu aufgebaut werden.
						// Das gleiche soll bei mehr als 500 unsortierten Einträgen passieren, sofern diese mindestens ein drittel so groß sind, wie die Anzahl
						// der sortierten Einträge.
						return removeContainerIDWithRebuild(containerID, randomAccessFile);
					}
					else {
						return removeContainerIDFast(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);
					}
				}
			}
		}

		/**
		 * Gibt zurück, ob eine Container-ID im Index enthalten ist
		 *
		 * @param containerID ContainerID die zu entfernen ist.
		 * @return True falls enthalten, false sonst
		 * @throws IOException IO-Fehler
		 */
		public boolean containsContainerID(long containerID) throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.readInt();
					final int numOfSortedEntries = randomAccessFile.readInt();

					return containsContainerID(containerID, randomAccessFile, numOfSortedEntries);
				}
			}
		}

		public int getMaxMedienID() throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					return randomAccessFile.readInt();
				}
			}
		}

		public void setMaxMedienID(int mediumID) throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "rw")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.writeInt(mediumID);
				}
			}
		}

		/**
		 * Markiert einen Container-Eintrag als ungültig
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @param numOfGaps          Anzahl ungültige Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDFast(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {

			boolean entryRemoved = removeContainerIDInSortedArea(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);

			if(!entryRemoved) {
				entryRemoved = removeContainerIDInUnsortedArea(containerID, randomAccessFile, numOfSortedEntries, numOfGaps);
			}
			return entryRemoved;
		}

		/**
		 * Markiert einen Container-Eintrag als ungültig
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean containsContainerID(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries) throws IOException {

			boolean found = containsContainerIDInSortedArea(containerID, randomAccessFile, numOfSortedEntries);

			if(!found) {
				found = containsContainerIDInUnsortedArea(containerID, randomAccessFile, numOfSortedEntries);
			}
			return found;
		}

		/**
		 * Löscht einen Container-Eintrag aus dem sortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean containsContainerIDInSortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries) throws IOException {
			int min = 0;
			int max = numOfSortedEntries - 1;
			int mid;
			while(max >= min) {

				mid = (max + min) / 2;
				randomAccessFile.seek(HEADER_SIZE + mid * ENTRY_SIZE); // zum Eintrag gehen
				final long idAtMid = randomAccessFile.readLong();

				if(containerID < idAtMid) {
					max = mid - 1;
				}
				else if(containerID > idAtMid) {
					min = mid + 1;
				}
				else {
					if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX) {
						return true;
					}
					break;
				}
			}
			return false;
		}

		/**
		 * Löscht einen Container-Eintrag aus dem unsortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean containsContainerIDInUnsortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries) throws IOException {
			randomAccessFile.seek(HEADER_SIZE + numOfSortedEntries * ENTRY_SIZE);
			while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
				final long id = randomAccessFile.readLong();
				if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX && id == containerID) {
					return true;
				}
			}
			return false;
		}

		/**
		 * Löscht einen Container-Eintrag aus dem sortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @param numOfGaps          Anzahl ungültige Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDInSortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {
			int min = 0;
			int max = numOfSortedEntries - SIZE_BYTE;
			int mid;
			while(max >= min) {

				mid = (max + min) / 2;
				randomAccessFile.seek(HEADER_SIZE + mid * ENTRY_SIZE); // zum Eintrag gehen
				final long idAtMid = randomAccessFile.readLong();

				if(containerID < idAtMid) {
					max = mid - SIZE_BYTE;
				}
				else if(containerID > idAtMid) {
					min = mid + SIZE_BYTE;
				}
				else {
					if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX) {
						randomAccessFile.seek(randomAccessFile.getFilePointer() - SIZE_BYTE);
						//gefunden
						randomAccessFile.writeByte(CONTAINER_DELETED_FROM_INDEX);
						randomAccessFile.seek(HEADER_POS_NUM_GAPS);
						randomAccessFile.writeInt(numOfGaps + 1);
						return true;
					}
					break;
				}
			}
			return false;
		}

		/**
		 * Löscht einen Container-Eintrag aus dem unsortierten Bereich
		 *
		 * @param containerID        ContainerID
		 * @param randomAccessFile   geöffnete Datei
		 * @param numOfSortedEntries Anzahl Sortierte Einträge
		 * @param numOfGaps          Anzahl ungültige Einträge
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDInUnsortedArea(
				final long containerID, final RandomAccessFile randomAccessFile, final int numOfSortedEntries, final int numOfGaps) throws IOException {
			randomAccessFile.seek(HEADER_SIZE + numOfSortedEntries * ENTRY_SIZE);
			while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
				final long id = randomAccessFile.readLong();
				if(randomAccessFile.readByte() != CONTAINER_DELETED_FROM_INDEX && id == containerID) {
					randomAccessFile.seek(randomAccessFile.getFilePointer() - SIZE_BYTE);
					randomAccessFile.writeByte(CONTAINER_DELETED_FROM_INDEX);
					randomAccessFile.seek(HEADER_POS_NUM_GAPS);
					randomAccessFile.writeInt(numOfGaps + 1);
					return true;
				}
			}
			return false;
		}

		/**
		 * Löscht einen Container-Eintrag aus der Datei und sortiert dabei den kompletten Datenbereich neu
		 *
		 * @param containerIdToRemove ContainerID
		 * @param randomAccessFile    geöffnete Datei
		 * @return True wenn der Container gelöscht wurde
		 * @throws IOException IO-Fehler
		 */
		private boolean removeContainerIDWithRebuild(
				final long containerIdToRemove, final RandomAccessFile randomAccessFile) throws IOException {

			_debug.fine("Baue Index neu auf");

			final ArrayList<IndexEntry> entryList = new ArrayList<IndexEntry>((int) (_file.length() / ENTRY_SIZE));
			boolean entryRemoved = false;
			// Alle Einträge einlesen...
			while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
				final long id = randomAccessFile.readLong();
				final byte type = randomAccessFile.readByte();
				if(type == CONTAINER_DELETED_FROM_INDEX) continue; // Gelöschte Einträge ignorieren
				if(id == containerIdToRemove) {
					// Den zu löschenden Container ebenso
					entryRemoved = true;
					continue;
				}
				entryList.add(new IndexEntry(id, type));
			}
			Collections.sort(entryList);
			// ...und jetzt die Datei neu schreiben
			randomAccessFile.setLength(HEADER_SIZE + entryList.size() * ENTRY_SIZE);
			randomAccessFile.seek(HEADER_POS_NUM_SORTED_ENTRIES);
			randomAccessFile.writeInt(entryList.size());
			randomAccessFile.writeInt(0);
			for(IndexEntry entry : entryList) {
				randomAccessFile.writeLong(entry.getContainerID());
				randomAccessFile.writeByte(entry.getType());
			}
			return entryRemoved;
		}

		@Override
		public String toString() {
			synchronized(_fileLock) {
				return _file.toString();
			}
		}

		/**
		 * Gibt alle ContainerIDs im Index zurück
		 *
		 * @return Liste mit ContainerIDs
		 * @throws IOException IO-Fehler
		 */
		public List<Long> getContainerIDs() throws IOException {
			synchronized(_fileLock) {
				try(RandomAccessFile randomAccessFile = new RandomAccessFile(_file, "r")) {
					final int version = randomAccessFile.readInt();
					if(version != 1) throw new IOException(_file + " hat unbekannte Dateiversion: " + version);
					randomAccessFile.seek(HEADER_SIZE);
					final ArrayList<Long> fileList = new ArrayList<Long>((int) ((randomAccessFile.length() - HEADER_SIZE) / ENTRY_SIZE));
					while(randomAccessFile.getFilePointer() < randomAccessFile.length()) {
						final long id = randomAccessFile.readLong();
						final byte type = randomAccessFile.readByte();
						if(type == CONTAINER_DELETED_FROM_INDEX) continue;
						fileList.add(id);
					}
					return fileList;
				}
			}
		}

		/** Hilfsklasse die einen einzelnen Dateneintrag speichert. Wird zum Sortieren benutzt. */
		private final static class IndexEntry implements Comparable {

			private long _containerID;

			private byte _type;

			public IndexEntry(long containerID, byte type) {
				this._containerID = containerID;
				this._type = type;
			}

			public int compareTo(Object o) {
				return ((Long) _containerID).compareTo(((IndexEntry) o).getContainerID());
			}

			public long getContainerID() {
				return _containerID;
			}

			public byte getType() {
				return _type;
			}

			@Override
			public String toString() {
				return "IndexEntry{" + "_containerID=" + _containerID + ", _type=" + _type + '}';
			}
		}
	}
}
