/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * Copyright 2010 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.index;

import de.bsvrz.ars.ars.persistence.index.result.DataTimeIndexEntry;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * Diese Klasse stellt einen DatenzeitIndex dar. Sie verhält sich ähnlich dem {@link ATimeMonotoneIndex}. Der Unterschied ist, dass die Folge der Eintraege
 * (min, max, min, max...) nicht monoton ist, weil es Rueckspruenge geben kann. Stattdessen sind die Eintraege nur nach dem min-Wert sortiert.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class DataTimeIndex {

	/** Dateiname, unter dem alle Indices dieser Art gespeichert werden */
	public static final String IDX_FILENAME = "_datatime.idx";

	/** Flag, mit dem signalisiert wird, dass ein Eintrag "keinen Wert" hat */
	public static final int NO_VALUE = -1;

	/** Modi, mit denen ein Index angelegt werden kann (siehe {@link DataTimeIndex#DataTimeIndex(int,String,byte[],IndexStartmode) } ) */
	public static enum IndexStartmode {

		APPEND_REUSE_LAST,
		ERASE_OLD
	}

	;

	public static final IndexStartmode APPEND_REUSE_LAST = IndexStartmode.APPEND_REUSE_LAST;

	public static final IndexStartmode ERASE_OLD = IndexStartmode.ERASE_OLD;

	/** Positionen einzelner Daten im Enty-Array */
	private static final int DT_MIN_POS = 0;

	private static final int DT_MAX_POS = 1;

	private static final int DI_MIN_POS = 2;

	private static final int DI_MAX_POS = 3;

	private static final int CONTID_POS = 4;

	private static final String ERROR_OPEN = "Die Indexdatei konnte nicht geöffnet werden. ";

	private static final String ERROR_CLOSE = "Die Indexdatei konnte nicht geschlossen werden. ";

	private static final String ERROR_ACCESS = "Fehler beim Zugriff auf die Indexdatei. ";

	/** Hilfsvariablen */
	private long biggestDTMaxInFile = NO_VALUE;

	private long filePositionReactivatedEntry = NO_VALUE;

	/** Current Entry Array - aktuelle Daten */
	private long currentEntry[] = new long[]{NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE};

	/**
	 * Erzeugt einen neuen Datenzeitindex.
	 * <p>
	 * Bei startmode==APPEND_REUSE_LAST wird der Ringpuffer mit den aktuellsten Daten aus einer vorhandenen Indexdatei gefüllt. Der Eintrag mit der größten
	 * ContainerId wird als currentEntry reaktiviert. Damit kann ein Container fortgesetzt und der reaktivierte Eintrag weiter aktualisiert werden.
	 * <p>
	 * Bei startmode==ERASE_OLD wird ein leerer Index erzeugt und die bisherige Indexdatei (sofern vorhanden) geloescht.
	 *
	 * @param bufferSize Anzahl der Einträge, die der die der Index im RAM halten soll
	 * @param basePath   Verzeichnis, in dem die Indexdatei liegt
	 * @param temp       Ein temporäres ByteArray, das zum Suchen des currentEntries verwendet wird
	 * @param startmode  {@link IndexStartmode#ERASE_OLD}: alte Indexdatei loeschen, {@link IndexStartmode#APPEND_REUSE_LAST}: anfuegen, letzten Eintrag
	 *                   weiterverwenden
	 *
	 * @throws IndexException Bei Problemem mit dem Dateizugriff
	 */
	public DataTimeIndex(int bufferSize, String basePath, byte temp[], IndexStartmode startmode) throws IndexException {
		this(bufferSize);
		if(startmode == IndexStartmode.ERASE_OLD) {
			File oldIndexFile = new File(basePath, IDX_FILENAME);
			if(oldIndexFile.exists() && !oldIndexFile.delete()) {
				throw new IndexException(
						"Fehler beim Neuanlegen eines Index: Alte Indexdatei '" + oldIndexFile.getAbsolutePath() + "' konnte nicht geloescht werden: "
				);
			}
		}
		else if(startmode == IndexStartmode.APPEND_REUSE_LAST) {
			if(new File(basePath, IDX_FILENAME).exists()) {
				RandomAccessFile raf = ifOpen(basePath);
				try {
					rbLoad(raf);
					long maxima[][] = getMaximaFromFile(raf, temp, 0, (int)raf.length());
					biggestDTMaxInFile = maxima[0][0];
					if(currentEntry[CONTID_POS] < maxima[2][CONTID_POS]) {
						currentEntry = maxima[2];
						filePositionReactivatedEntry = maxima[1][0];
					}
					rbExtractMaxima();
				}
				catch(Exception e) {
					throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
				}
				finally {
					ifClose(raf);
				}
			}
		}
	}


	/**
	 * Erzeugt einen neuen DatenZeitIndex. Eine eventuell vorhandene Datei wird weiterbenutzt, ein weiterbenutzen des currentEntries ist nicht möglich.
	 *
	 * @param bufferSize Anzahl der Einträge, die der die der Index im RAM halten soll
	 */
	public DataTimeIndex(int bufferSize) {
		assert bufferSize > 0 : "bufferSize muss größer 0 sein.";
		bufferSize = bufferSize <= 0 ? 1 : bufferSize;
		rbInit(bufferSize);
	}


	////////////////////////////// Methoden auf den Current*-Werten ////////////////////////////
	/**
	 * Aktualisiert den Max-Wert des aktuellsten Eintrages. Der Min-Wert und die ContainerID bleiben unverändert
	 *
	 * @param newDTMax neuer Wert für DatenZeitIndex-Max
	 * @param newDIMax neuer Wert für DatenIndexIndex-Max
	 *
	 * @throws IndexException Wenn der neue Wert kleiner als der dazugehörige Min-Wert ist oder wenn keinen currentEntry gibt
	 */
	public void updateMax(long newDTMax, long newDIMax) throws IndexException {
		if(currentEntry[CONTID_POS] == NO_VALUE) throw new IndexException("Update, obwohl noch kein Eintrag hinzugefügt wurde");
		if(newDTMax < currentEntry[DT_MIN_POS]) throw new IndexException("neuer DTMax-Wert ist kleiner als DTMin-Wert");
		if(newDIMax < currentEntry[DI_MIN_POS]) throw new IndexException("neuer DIMax-Wert ist kleiner als DIMin-Wert");

		currentEntry[DT_MAX_POS] = newDTMax;
		currentEntry[DI_MAX_POS] = newDIMax;
	}


	/**
	 * Liefert die aktuelle Container ID. Wenn es noch keine Aktuelle Container ID gibt, wird -1 zurückgeliefert
	 *
	 * @return die aktuelle Container ID oder -1, wenn es keinen aktuellen Entry gibt.
	 */
	public long getCurrentCID() {
		return currentEntry[CONTID_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Min-Wert
	 *
	 * @return Min-Wert des aktuellen Eintrags
	 */
	public long getCurrentDTMin() {
		return currentEntry[DT_MIN_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Max-Wert
	 *
	 * @return Max-Wert des aktuellen Eintrags
	 */
	public long getCurrentDTMax() {
		return currentEntry[DT_MAX_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Min-Wert des DatenIndex
	 *
	 * @return Min-Wert des DatenIndex des aktuellen Eintrags
	 */
	public long getCurrentDIMin() {
		return currentEntry[DI_MIN_POS];
	}


	/**
	 * Liefert vom aktuellen Eintrag den Max-Wert des DatenIndex
	 *
	 * @return Max-Wert des DatenIndex des aktuellen Eintrags
	 */
	public long getCurrentDIMax() {
		return currentEntry[DI_MAX_POS];
	}


	/**
	 * Anzahl der Einträge, die dieser Index im RAM halten kann
	 *
	 * @return Anzahl der Einträge
	 */
	public int capacity() {
		return rbCapacity();
	}


	/**
	 * Anzahl der Einträge, die dieser Index im RAM hält
	 *
	 * @return Anzahl der Einträge
	 */
	public int entries() {
		return rbEntries();
	}


	/**
	 * Fügt dem Index einen neuen Eintrag hinzu. Dabei wird der derzeitig aktuelle Eintrag abgeschlossen. Führt dies dazu, dass keine weiteren freien Plätze im
	 * Cache vorhanden sind, werden die Änderungen auf Platte geschrieben. Wenn der Cache noch vollständig leer ist, werden die aktuellsten Daten aus der
	 * Index-Datei geladen.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dTMin    DatenZeit Min-Wert
	 * @param dTMax    DatenZeit Max-Wert
	 * @param dIMin    DatenIndex Min-Wert
	 * @param dIMax    DatenIndex Max-Wert
	 * @param contId   Container ID
	 *
	 * @throws IndexException Bei Problemem mit der Index-Datei
	 */
	public void add(String basePath, byte[] temp, long dTMin, long dTMax, long dIMin, long dIMax, long contId) throws IndexException {
		checkForAdd(basePath, temp, dTMin, dTMax, dIMin, dIMax, contId);

		if(rbMustBeSaved()) flush(basePath);

		if(currentEntry[DT_MIN_POS] == NO_VALUE) {
			if(biggestDTMaxInFile == NO_VALUE && new File(basePath, IDX_FILENAME).exists()) {
				RandomAccessFile raf = ifOpen(basePath);
				rbLoad(raf);
				ifClose(raf);
			}
		}
		else {
			currentEntryToBuffer(basePath, temp);
			if(biggestDTMaxInFile == NO_VALUE) { // Wir wissen nichts über die Datei
				if(new File(basePath, IDX_FILENAME).exists()) {
					RandomAccessFile raf = ifOpen(basePath);
					try {
						if(raf.length() > 0) {
							int cachedEntries = (rbFirstUnsaved == NO_VALUE) ? rbEntries() : rbFirstUnsaved;
							if(((int)raf.length() - IndexEntryTime.ENTRY_LEN * cachedEntries) > 0) {
								biggestDTMaxInFile = getMaximaFromFile(raf, temp, 0, (int)raf.length() - IndexEntryTime.ENTRY_LEN * cachedEntries)[0][0];
							}
						}
					}
					catch(IOException e) {
						throw new IndexException(e);	// Sollte nie auftreten
					}
					finally {
						ifClose(raf);
					}
				}
			}
		}

		currentEntry[DT_MIN_POS] = dTMin;
		currentEntry[DT_MAX_POS] = dTMax;
		currentEntry[DI_MIN_POS] = dIMin;
		currentEntry[DI_MAX_POS] = dIMax;
		currentEntry[CONTID_POS] = contId;
	}

	/**
	 * Prüft, ob dem Index ein neuer Eintrag hinzugefügt werden könnte.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dTMin    DatenZeit Min-Wert
	 * @param dTMax    DatenZeit Max-Wert
	 * @param dIMin    DatenIndex Min-Wert
	 * @param dIMax    DatenIndex Max-Wert
	 * @param contId   Container ID
	 *
	 * @throws IndexException Bei Problemem mit der Index-Datei
	 */
	public void checkForAdd(String basePath, byte[] temp, long dTMin, long dTMax, long dIMin, long dIMax, long contId) throws IndexException {
		if(dIMin < 0 || dIMax < 0 || dTMin < 0 || dTMax < 0 || contId <= 0 || dIMin > dIMax || dTMin > dTMax) {
			throw new CorruptIndexException(
					"DatenzeitIndex: neuer Eintrag ungültig: (" + "dIMin=" + dIMin + ", dIMax=" + dIMax + ", dTMin=" + dTMin + ", dTMax=" + dTMax + ", contID="
					+ contId + ")"
			);
		}
		if(getCurrentCID() >= contId) {
			throw new CorruptIndexException("DatenzeitIndex: ContainerId (" + contId + ") des neuen Eintrags ungültig: currentCId=" + getCurrentCID());
		}
	}


	/**
	 * Liefert eine Matrix mit DTMin/Max, DIMin/Max und ContainerIds, deren DTMin/Max-Werte zwischen dTMin und dTMax liegen. Wird diese Methode aufgerufen wenn der
	 * Cache noch leer ist, werden die aktuellsten Daten aus der Datei nachgeladen.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Einlesen/Schreiben der Index-Datei verwendet wird
	 * @param dTMin    Min-Wert der Zeitspanne
	 * @param dTMax    Max-Wert der Zeitspanne
	 *
	 * @return IndexResult-Objekt, das die Index innerhalb der gesuchten Zeitspanne enthält (DT und DI Werte, keine AT Werte)
	 *
	 * @throws IndexException Bei Problemem mit der IndexDatei
	 */
	public IndexResult<DataTimeIndexEntry> getContainerID(String basePath, byte[] temp, long dTMin, long dTMax) throws IndexException {
		IndexResult<DataTimeIndexEntry> ir = IndexResult.newDataTimeIndexResult();
		RandomAccessFile raf = null;

		if(currentEntry[DT_MIN_POS] == NO_VALUE) {
			try {
				raf = ifOpen(basePath);
				rbLoad(raf);
			}
			catch(IndexException e) {
				return ir;
			}
		}

		if(biggestDTMaxInFile == NO_VALUE) {
			if(new File(basePath, IDX_FILENAME).exists()) {
				if(raf == null) raf = ifOpen(basePath);
				try {
					if(raf.length() > 0) {
						int cachedEntries = (rbFirstUnsaved == NO_VALUE) ? rbEntries() : rbElementsInbetween(rbFirstUnsaved, rbSmallestEntry, rbEntries());
						long maxima[][] = getMaximaFromFile(raf, temp, 0, (int)raf.length() - IndexEntryTime.ENTRY_LEN * cachedEntries);
						biggestDTMaxInFile = maxima[0][0];
					}
				}
				catch(IOException e) {
					throw new IndexException(e);	// Sollte nie auftreten
				}
			}
		}

		boolean firstContainerFound = false;
		long previousContainerId = -1;
		long previousDataTimeMinimum = -1;
		long previousDataTimeMaximum = -1;
		long previousDataIndexMinimum = -1;
		long previousDataIndexMaximum = -1;

		if(biggestDTMaxInFile != NO_VALUE && dTMin <= biggestDTMaxInFile) {
			if(raf == null) raf = ifOpen(basePath);
			long[] previousContainerInfos = linearFileSearch(raf, temp, ir, dTMin, dTMax);
			if(ir.size() >0) {
				firstContainerFound = true;
			}
			if(previousContainerInfos != null) {
				previousContainerId = previousContainerInfos[0];
				previousDataTimeMinimum = previousContainerInfos[1];
				previousDataTimeMaximum = previousContainerInfos[2];
				previousDataIndexMinimum = previousContainerInfos[3];
				previousDataIndexMaximum = previousContainerInfos[4];
			}
		}
		ifClose(raf);

		int previousContainerIndex = -1;
X:
		for(int i = (rbGetFirstUnsaved() <= 0 || dTMin >= biggestDTMaxInFile) ? 0 : rbGetFirstUnsaved(); i < rbEntries(); i++) {
			if(overlap(rbGetDTMin(i), rbGetDTMax(i), dTMin, dTMax)) {
				for(int j = ir.size() - 1; j >= 0; j--) {
					if(ir.get(j).getContainerId() == rbGetCID(i)) continue X;
				}
				long dataTimeMinimum = rbGetDTMin(i);
				ir.add(new DataTimeIndexEntry(rbGetCID(i), rbGetDIMin(i), rbGetDIMax(i), dataTimeMinimum, rbGetDTMax(i)));
				firstContainerFound = true;
			}
			else if(!firstContainerFound) {
				long dataTimeMaximum = rbGetDTMax(i);
				if(dataTimeMaximum < dTMin && dataTimeMaximum >= previousDataTimeMaximum) {
					previousContainerIndex = i;
					previousDataTimeMaximum = dataTimeMaximum;
				}
			}
		}
		if(previousContainerIndex != -1) {
			previousContainerId = rbGetCID(previousContainerIndex);
			previousDataTimeMinimum = rbGetDTMin(previousContainerIndex);
			// previousDataTimeMaximum ist schon gesetzt
			previousDataIndexMinimum = rbGetDIMin(previousContainerIndex);
			previousDataIndexMaximum = rbGetDIMax(previousContainerIndex);
		}
Y:
		if(currentEntry[DT_MIN_POS] != NO_VALUE) {
			if(overlap(currentEntry[DT_MIN_POS], currentEntry[DT_MAX_POS], dTMin, dTMax)) {
				firstContainerFound = true;
				if(ir.size() == 0 || (ir.get(ir.size() - 1).getDataTimeMinimum() < currentEntry[DT_MIN_POS] && filePositionReactivatedEntry == NO_VALUE)) {
					ir.add(new DataTimeIndexEntry(currentEntry[CONTID_POS], currentEntry[DI_MIN_POS], currentEntry[DI_MAX_POS], currentEntry[DT_MIN_POS], currentEntry[DT_MAX_POS]));
					break Y;
				}
				for(int j = ir.size() - 1; j >= 0; j--) {
					if(ir.get(j).getContainerId() == currentEntry[CONTID_POS]) break Y;
					if(ir.get(j).getDataTimeMinimum() <= currentEntry[DT_MIN_POS]) {
						ir.insert(
								j + 1,
								new DataTimeIndexEntry(currentEntry[CONTID_POS], currentEntry[DI_MIN_POS], currentEntry[DI_MAX_POS], currentEntry[DT_MIN_POS], currentEntry[DT_MAX_POS])
						);
						break Y;
					}
				}
				ir.insert(0, new DataTimeIndexEntry(currentEntry[CONTID_POS], currentEntry[DI_MIN_POS], currentEntry[DI_MAX_POS], currentEntry[DT_MIN_POS], currentEntry[DT_MAX_POS]));
			}
			else if(!firstContainerFound) {
				long dataTimeMaximum = currentEntry[DT_MAX_POS];
				if(dataTimeMaximum < dTMin && dataTimeMaximum >= previousDataTimeMaximum) {
					previousContainerId = currentEntry[CONTID_POS];
					previousDataTimeMinimum = currentEntry[DT_MIN_POS];
					previousDataTimeMaximum = dataTimeMaximum;
					previousDataIndexMinimum = currentEntry[DI_MIN_POS];
					previousDataIndexMaximum = currentEntry[DI_MAX_POS];
				}
			}

		}
		// Wenn kein passender Container gefunden wurde, dann den Container, der am nächsten dran ist auswählen
		if(!firstContainerFound && previousContainerId != -1) {
			ir.add(new DataTimeIndexEntry(previousContainerId, previousDataIndexMinimum, previousDataIndexMaximum, previousDataTimeMinimum, previousDataTimeMaximum));
		}
		return ir;
	}


	/**
	 * Liefert den kleinsten DT Min Wert, der jemals in diesem Index gespeichert wird.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 *
	 * @return Kleinste Datenzeit oder -1 falls keine vorhanden
	 *
	 * @throws IndexException Bei Zugriffsproblemen mit der IndexDatei
	 */
	public long getAbsolutDTMin(String basePath) throws IndexException {
		RandomAccessFile raf = null;
		File f = new File(basePath, IDX_FILENAME);

		if(f.length() == 0) {
			if(currentEntry[DT_MIN_POS] == NO_VALUE) {
				// Fall 1.1: Datei leer, Ram leer           => Leere Antwort
				return NO_VALUE;
			}
			else {
				// Fall 1.2: Datei leer, Ram hält Daten     => Antwort ist ältestes Element im Ram
				if(rbEntries() > 0 && rbGetDTMin(0) < currentEntry[DT_MIN_POS]) {
					return rbGetDTMin(0);
				}
				else {
					return currentEntry[DT_MIN_POS];
				}
			}
		}
		else {
			if(currentEntry[DT_MIN_POS] == NO_VALUE) {
				// Fall 2.1: Datei hält Daten, Ram leer     => Daten laden. Antwort ist ältestes Element im Ram, wenn
				//                                             alle Daten ins Ram geladen werden konnten. Ansonsten wie 2.2 behandeln
				raf = ifOpen(basePath);
				rbLoad(raf);
				if(f.length() == rbEntries() * rbEntryLength()) {
					ifClose(raf);
					return rbGetDTMin(0);
				}
			}
			//    Fall 2.2: Daten und Ram enthält Daten     => Ältestes Element aus der Datei nehmen
			if(raf == null) raf = ifOpen(basePath);
			byte tmp[] = new byte[rbEntryLength()];
			int bytesRead = -1;
			try {
				raf.seek(0);
				bytesRead = raf.read(tmp, 0, tmp.length);
			}
			catch(IOException e) {
			}
			finally {
				ifClose(raf);
			}

			if(bytesRead == rbEntryLength()) {
				if(currentEntry[DT_MIN_POS] == NO_VALUE) {
					return IndexEntryTime.getDTMin(tmp, 0);
				}
				else {
					return (IndexEntryTime.getDTMin(tmp, 0) < currentEntry[DT_MIN_POS]) ? IndexEntryTime.getDTMin(tmp, 0) : currentEntry[DT_MIN_POS];
				}
			}

			return NO_VALUE;
		}
	}

	/**
	 * Schließt den Index und speichert alle geänderten Daten auf Platte. Nach dem Aufruf von close() darf die Index-Instanz nicht mehr verwendet werden.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 * @param temp     Puffer, der beim Schreiben der Index-Datei verwendet wird
	 *
	 * @throws IndexException
	 */
	public void closeIndex(String basePath, byte[] temp) throws IndexException {
		if(rbMustBeSaved()) flush(basePath);
		if(currentEntry[DT_MIN_POS] != NO_VALUE) currentEntryToBuffer(basePath, temp);
		flush(basePath);
	}


	/**
	 * Gibt den Inhalt des ByteBuffers als String aus, inkl Angaben über Kapazität, Anzahl der Einträge und des CurrentEntries.
	 *
	 * @return ByteBuffer als String
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("capacity: " + rbCapacity() + ", entries: " + rbEntries());
		sb.append(", rbFirstUnsaved: " + rbFirstUnsaved);
		sb.append(", filePositionReactivatedEntry: " + filePositionReactivatedEntry);
		sb.append(
				", current: (" + currentEntry[DT_MIN_POS] + "," + currentEntry[DT_MAX_POS] + "," + currentEntry[DI_MIN_POS] + "," + currentEntry[DI_MAX_POS]
				+ "," + currentEntry[CONTID_POS] + ") ringbuffer: "
		);
		return sb.toString() + rbToString();
	}


	// ///////////////////////////// Private Helpermethoden /////////////////////////////////
	private void currentEntryToBuffer(String basePath, byte[] temp) throws IndexException {
		if(filePositionReactivatedEntry != NO_VALUE) {
			updateCurrentEntryInFile(basePath);
		}
		else if(currentEntry[DT_MIN_POS] >= rbGetDTMin(0)) {
			rbAddCurrentEntry();
		}
		else {
			insertIntoFile(basePath, temp);
		}
	}


	private void flush(String basePath) throws IndexException {
		if(rbHasUnsavedData()) {
			try {
				byte[] bytesToWrite;

				int startByte =  ((rbSmallestEntry + rbFirstUnsaved) % rbCapacity()) * IndexEntryTime.ENTRY_LEN;
				int endByte = rbBiggestEntry * IndexEntryTime.ENTRY_LEN + IndexEntryTime.ENTRY_LEN;

				if(startByte >= endByte) {
					bytesToWrite = new byte[rbBuffer.length - startByte + endByte];
					final int tmp = rbBuffer.length - startByte;
					System.arraycopy(rbBuffer, startByte, bytesToWrite, 0, tmp);   // Hinterer Teil
					System.arraycopy(rbBuffer, 0, bytesToWrite, tmp, endByte);     // Vorderer Teil
					startByte = 0;
					endByte = bytesToWrite.length;
				}
				else {
					bytesToWrite = rbBuffer;
				}

				if(rbCacheChanged()) {
					RandomAccessFile raf = ifOpen(basePath);
					try {
						long entry[] = rbGetFileAnchor();
						byte buffer[] = new byte[(raf.length() > rbCapacityInBytes()) ? rbCapacityInBytes() : (int)raf.length()];
						raf.seek(raf.length() - buffer.length);
						raf.read(buffer);
						int idx = locateEntry(buffer, entry);
						if(idx > 0) raf.setLength(raf.length() - (idx * IndexEntryTime.ENTRY_LEN));
						raf.seek(raf.length());
						raf.write(bytesToWrite, startByte, endByte - startByte);
					}
					finally {
						ifClose(raf);
					}
				}
				else {
					final FileOutputStream os = new FileOutputStream(new File(basePath, IDX_FILENAME), true);
					try {
						os.write(bytesToWrite, startByte, endByte - startByte);
					}
					finally {
						os.close();
					}
				}

				rbFirstUnsaved = NO_VALUE;
			}
			catch(Exception e) {
				throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
			}
		}
	}


	private void insertIntoFile(String basePath, byte temp[]) throws IndexException {
		RandomAccessFile raf = null;
		try {
			raf = ifOpen(basePath);
			byte entry[] = new byte[rbEntryLength()];
			IndexEntryTime.setEntry(entry, 0, currentEntry);

			int start = 0;
			int end = (int)raf.length();

			if(end == 0) {
				raf.write(entry); // Leere Datei, kein umkopieren
			}
			else {
				int position = (int)binaryFileSearch(raf, temp, currentEntry[DT_MIN_POS]);

				while(start >= 0) {
					start = end - temp.length;
					if(start <= position) {
						raf.seek(position);
						int bytesRead = raf.read(temp, 0, (end - position));
						raf.seek(position);
						raf.write(entry);
						if(bytesRead > 0) raf.write(temp, 0, bytesRead);
						if(currentEntry[DT_MAX_POS] > biggestDTMaxInFile) biggestDTMaxInFile = currentEntry[DT_MAX_POS];
						break;
					}
					else {
						raf.seek(start);
						int bytesRead = raf.read(temp);
						raf.seek(start + entry.length);
						raf.write(temp, 0, bytesRead);
					}
					end = start;
				}
			}
		}
		catch(Exception e) {
			throw new CorruptIndexException("Probleme beim Umkopieren der Indexdatei. Sie ist nun sehr wahrscheinlich beschädigt: " + e.toString(), e);
		}
		finally {
			ifClose(raf);
		}
	}


	/**
	 * Ermittelt mit einer binären Suche die Position, an der der Key1 untergebracht werden kann. Wenn es schon mehrere gleiche Key1 gibt, wird ein beliebiger
	 * davon zurück gegeben.
	 *
	 * @param raf  Eine Referenz auf die Datei
	 * @param temp Ein Temp Array, zum Einlesen der Datei. Je größer umso besser
	 * @param key  der gesuchte Key1
	 *
	 * @return die bytePosition an der der Key eingefügt werden muss
	 *
	 * @throws IndexException
	 * @throws IOException
	 */
	private long binaryFileSearch(RandomAccessFile raf, byte[] temp, long key) throws IndexException, IOException {
		assert temp.length >= rbEntryLength() : "Puffer kleiner als Länge eines Eintrags";
		if(temp.length < rbEntryLength()) temp = new byte[(1024 / rbEntryLength()) * rbEntryLength()];

		int useableArea = (temp.length / rbEntryLength()) * rbEntryLength();
		int numberOfChunks = ((int)raf.length() + useableArea - 1 )/ useableArea;

		// Denn richtigen Abschnitt in der Datei finden
		int fMid = 0, fLow = 0, bytesRead = -1;
		int fHigh = numberOfChunks - 1;
		while(fLow <= fHigh) {
			fMid = (fLow + fHigh) >> 1;
			raf.seek(fMid * useableArea);
			bytesRead = raf.read(temp, 0, useableArea);

			int cmp = containsKey(temp, bytesRead / IndexEntryTime.ENTRY_LEN, key);
			if(cmp < 0) {
				fHigh = fMid - 1;
			}
			else if(cmp > 0) {
				fLow = fMid + 1;
			}
			else {
				break;
			}
		}

		// und jetzt den richtigen Eintrag im eingelesenen Chunk suchen
		int low = 0, mid = 0;
		int high = bytesRead / IndexEntryTime.ENTRY_LEN - 1;
		int max = high;
		while(low <= high) {
			mid = (low + high) >> 1;
			long midVal = IndexEntryTime.getDTMin(temp, mid);

			if(midVal < key) {
				low = mid + 1;
			}
			else if(midVal > key) {
				high = mid - 1;
			}
			else {
				while(midVal == key ){
					mid++;
					if(mid>max) break;
					midVal = IndexEntryTime.getDTMin(temp, mid);
				}
				return mid * IndexEntryTime.ENTRY_LEN + fMid * useableArea;
			}
		}
		return low * IndexEntryTime.ENTRY_LEN + fMid * useableArea;
	}


	/**
	 * Sucht mit Hilfe eine linearen Suche auf der Indexdatei Container der Datenzeiten sich mit dem übergebenen Bereich schneiden und speicher diese im
	 * übergebenen Indexresult.
	 * Informationen zu einem vorhergehenden Container des Maximum am nächsten am Minimum des Zeitbereichs liegt, werden in einem Array zurückgeben, falls
	 * vorhanden.
	 *
	 * @param raf  Eine Referenz auf die Datei
	 * @param temp Ein Temp Array, zum Einlesen der Datei. Je größer umso besser
	 * @param ir Speichert die gefundenen Container
	 * @param min Minimum des gesuchten Zeitbereichs
	 * @param max Maximum des gesuchten Zeitbereichs
	 * @return Informationen zu einem vorhergehenden Container des Maximum am nächsten am Minimum des Zeitbereichs liegt.
	 * @throws IndexException
	 */
	private long[] linearFileSearch(RandomAccessFile raf, byte[] temp, IndexResult<DataTimeIndexEntry> ir, long min, long max) throws IndexException {
		assert temp.length >= rbEntryLength() : "Puffer kleiner als Länge eines Eintrags";

		if(temp.length < rbEntryLength()) temp = new byte[(1024 / rbEntryLength()) * rbEntryLength()];

		boolean firstContainerFound = false;
		long previousContainerId = -1;
		long previousDataTimeMinimum = -1;
		long previousDataTimeMaximum = -1;
		long previousDataIndexMinimum = -1;
		long previousDataIndexMaximum = -1;

		try {
			int useableArea = (temp.length / rbEntryLength()) * rbEntryLength();
			int chunks = ((int)raf.length() / useableArea);

			for(int i = 0; i <= chunks; i++) {
				raf.seek(i * useableArea);
				int bytesRead = raf.read(temp, 0, useableArea);

				int entries = bytesRead / IndexEntryTime.ENTRY_LEN;
				for(int j = 0; j < entries; j++) {
					if(overlap(IndexEntryTime.getDTMin(temp, j), IndexEntryTime.getDTMax(temp, j), min, max)) {
						ir.add(
								new DataTimeIndexEntry(IndexEntryTime.getCID(temp, j), IndexEntryTime.getDIMin(
																temp, j
														), IndexEntryTime.getDIMax(temp, j), IndexEntryTime.getDTMin(temp, j), IndexEntryTime.getDTMax(temp, j))
						);
					}
					else if(!firstContainerFound) {
						long dataTimeMaximum = IndexEntryTime.getDTMax(temp, j);
						if(dataTimeMaximum < min && dataTimeMaximum >= previousDataTimeMaximum) {
							previousContainerId = IndexEntryTime.getCID(temp, j);
							previousDataTimeMinimum = IndexEntryTime.getDTMin(temp, j);
							previousDataTimeMaximum = dataTimeMaximum;
							previousDataIndexMinimum = IndexEntryTime.getDIMin( temp, j);
							previousDataIndexMaximum = IndexEntryTime.getDIMax(temp, j);
						}
					}
				}
			}
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
		if(previousContainerId != -1) {
			return new long[]{previousContainerId, previousDataTimeMinimum, previousDataTimeMaximum, previousDataIndexMinimum, previousDataIndexMaximum};
		}
		return null;
	}


	private static int locateEntry(byte buffer[], long entry[]) {
		int elements = buffer.length / IndexEntryTime.ENTRY_LEN;
		for(int i = 0; i < elements; i++) {
			final long dtMin = IndexEntryTime.getDTMin(buffer, i);
			final long currentDtMin = entry[DT_MIN_POS];
			if(dtMin > currentDtMin || (dtMin == currentDtMin && IndexEntryTime.getCID(buffer, i) == entry[CONTID_POS])) {
				return elements - i;
			}
		}
		return -1;
	}


	private void updateCurrentEntryInFile(String basePath) throws IndexException {
		assert filePositionReactivatedEntry != NO_VALUE : "updateCurrentEntry(), obwohl es keinen wiederaufgenommen Eintrag gibt";

		RandomAccessFile raf = null;
		try {
			raf = ifOpen(basePath);
			raf.seek(filePositionReactivatedEntry);
			byte temp[] = new byte[rbEntryLength()];
			IndexEntryTime.setEntry(temp, 0, currentEntry);
			raf.write(temp);
			filePositionReactivatedEntry = NO_VALUE;
			rbFirstUnsaved = NO_VALUE;
			rbLoad(raf);
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
		finally {
			ifClose(raf);
		}
	}


	/**
	 * @param raf       Random Access File
	 * @param temp      Temp. Buffer
	 * @param fileStart Erstes Byte, ab den gelesen werden soll
	 * @param fileEnd   Letztes Byte, bis zu dem gelesen werden soll
	 *
	 * @return Matrix:  [0][0] = Größter DTMax-Wert [1][0] = erstes Byte des Eintrags mit der größten CID [2][]  = der Maximaeintrag selbst
	 *
	 * @throws IOException
	 */
	private long[][] getMaximaFromFile(RandomAccessFile raf, byte[] temp, int fileStart, int fileEnd) throws IOException {
		assert raf != null : "Kein RandomAccessfile";
		assert temp != null && temp.length >= rbEntryLength() : "temp nicht vorhanden oder nicht groß genug";
		assert fileStart <= fileEnd : "fileStart muss kleienr als FileEnd sein";

		long maxima[][] = {
				{NO_VALUE}, {NO_VALUE}, {NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE, NO_VALUE}
		};

		if(raf.length() <= 0) return maxima;

		// byteBuffer in benutzbare Bereiche einteilen
		int useableArea = ((int)temp.length / rbEntryLength()) * rbEntryLength();
		int chunks = ((int)(fileEnd - fileStart) / useableArea) + 1;

		int offset = fileStart;
		for(int i = 0; i < chunks; i++) {
			raf.seek(offset);
			int bytesRead = raf.read(temp, 0, useableArea);
			int entries = bytesRead / rbEntryLength();
			for(int j = 0; j < entries && (i * useableArea + fileStart + j * rbEntryLength()) < fileEnd; j++) {
				if(IndexEntryTime.getDTMax(temp, j) > maxima[2][DT_MAX_POS]) // größter DTMax
				{
					maxima[0][0] = IndexEntryTime.getDTMax(temp, j);
				}

				if(IndexEntryTime.getCID(temp, j) > maxima[2][CONTID_POS]) { // Eintrag mit größter CID
					maxima[2] = IndexEntryTime.getEntry(temp, j);
					maxima[1][0] = (i * useableArea) + fileStart + j * rbEntryLength();
				}
			}
			offset += bytesRead;
		}
		return maxima;
	}


	private int containsKey(byte[] array, int cnt, long key) {
		return key < IndexEntryTime.getDTMin(array, 0) ? -1 : (key >= IndexEntryTime.getDTMin(array, cnt - 1) ? 1 : 0);
	}


	private boolean overlap(long firstMin, long firstMax, long secondMin, long secondMax) {
		return (secondMin <= firstMax && secondMin >= firstMin) || (secondMax <= firstMax && secondMax >= firstMin)
		       || (firstMin <= secondMax && firstMin >= secondMin) || (firstMax <= secondMax && firstMax >= secondMin);
	}


	// ///////////////////////////////////////////////////////////////////////
	// IndexFile
	private RandomAccessFile ifOpen(String basePath) throws IndexException {
		RandomAccessFile raf = null;

		try {
			raf = new RandomAccessFile(new File(basePath, IDX_FILENAME), "rw");
			return raf;
		}
		catch(IOException e) {
			throw new IndexException(ERROR_OPEN + e.getMessage(), e);
		}
	}


	private void ifClose(RandomAccessFile raf) throws IndexException {
		if(raf != null) {
			try {
				raf.close();
			}
			catch(IOException e) {
				throw new IndexException(ERROR_CLOSE + e.getMessage(), e);
			}
		}
	}

	// ///////////////////////////////////////////////////////////////////////
	// Ab hier ist alles, was den RingPuffer betrifft. Oberhalb dieses
	// Kommentars dürfen nur die rp* Methoden verwendet werden, nicht
	// aber die Felder
	private byte rbBuffer[] = null;

	private int rbSmallestEntry = 0;

	private int rbBiggestEntry = NO_VALUE;

	private int rbFirstUnsaved = NO_VALUE;

	private boolean rbFileupdateNecessary = false;


	private void rbInit(int entries) {
		rbBuffer = new byte[entries * IndexEntryTime.ENTRY_LEN];
	}


	private void rbLoad(RandomAccessFile raf) throws IndexException {
		try {
			raf.seek(raf.length() > rbCapacityInBytes() ? raf.length() - rbCapacityInBytes() : 0);
			int numRead = raf.read(rbBuffer);
			if(numRead > 0 && numRead % rbEntryLength() != 0) throw new IndexException("Probleme beim Einlesen der Indexe. Ein Eintrag war unvollständig");

			if(numRead > 0) {
				rbSmallestEntry = 0;
				rbBiggestEntry = numRead / rbEntryLength() - 1;
			}
			else {
				rbSmallestEntry = 0;
				rbBiggestEntry = -1;
			}
		}
		catch(Exception e) {
			throw new IndexException(ERROR_ACCESS + e.getMessage(), e);
		}
	}


	/**
	 * Verändert die Größe des RingBuffers. Beim Verkleinern werden nicht gespeicherte Einträge auf Platte geschrieben
	 *
	 * @param newSize
	 * @param basePath
	 *
	 * @throws IndexException
	 */
	public void resize(int newSize, String basePath) throws IndexException {
		assert newSize > 0 : "Neue Größe kleiner/gleich 0";
		newSize = newSize > 0 ? newSize : 1;

		if(rbCapacity() == newSize)             // Kein Umkopieren nötig, da gleiche Größe
		{
			return;
		}
		if(rbEntries() == 0) {                  //  Keine Daten -> nur neuen Buffer anlegen
			rbBuffer = new byte[newSize * IndexEntryTime.ENTRY_LEN];
			return;
		}
		if(newSize > rbCapacity()) {                 // vergrößern
			if(rbSmallestEntry <= rbBiggestEntry) { // einfaches umkopieren reicht aus
				byte oldBuffer[] = rbBuffer;
				int start = rbBytePos(0);
				int length = rbBytePos(rbEntries() - 1) - start + IndexEntryTime.ENTRY_LEN;
				rbBuffer = new byte[newSize * IndexEntryTime.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, start, length);
			}
			else {                                // Daten überlappen Grenze. Neu anordnen und Zeiger anpassen
				byte oldBuffer[] = rbBuffer;
				int num = rbEntries();
				int start = rbBytePos(0);
				int length = rbCapacityInBytes() - start;
				rbBuffer = new byte[newSize * IndexEntryTime.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				start = length;
				length = rbBiggestEntry * IndexEntryTime.ENTRY_LEN + IndexEntryTime.ENTRY_LEN;
				System.arraycopy(oldBuffer, 0, rbBuffer, start, length);
				rbBiggestEntry = num - 1;
				rbSmallestEntry = 0;
			}
		}
		else {                                    // Verkleinern
			// Speichern, wenn es ungesicherte Daten gibt und diese mehr Platz belegen, als im neuen Puffer zur Verfügung stehen
//            if (rbHasUnsavedData() && ((rbCapacity() + rbBiggestEntry - rbFirstUnsaved) % rbCapacity()) > newSize)
			if(rbHasUnsavedData() && rbEntries() - rbFirstUnsaved > newSize) flush(basePath);
			int firstToCopy = (newSize > rbEntries()) ? rbSmallestEntry : (rbEntries() + rbBiggestEntry + 1 - newSize) % rbEntries();

			if(firstToCopy < rbBiggestEntry) {
				byte oldBuffer[] = rbBuffer;
				int start = firstToCopy * IndexEntryTime.ENTRY_LEN;
				int length = rbBytePos(rbEntries() - 1) - start + IndexEntryTime.ENTRY_LEN;
				rbBuffer = new byte[newSize * IndexEntryTime.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				if(rbFirstUnsaved != NO_VALUE) rbFirstUnsaved = (length / IndexEntryTime.ENTRY_LEN - 1) - (rbBiggestEntry - firstToCopy);
				rbSmallestEntry = 0;
				rbBiggestEntry = length / IndexEntryTime.ENTRY_LEN - 1;
			}
			else {
				byte oldBuffer[] = rbBuffer;
				int start = firstToCopy * IndexEntryTime.ENTRY_LEN;
				int length = rbCapacityInBytes() - start;
				rbBuffer = new byte[newSize * IndexEntryTime.ENTRY_LEN];
				System.arraycopy(oldBuffer, start, rbBuffer, 0, length);
				start = length;
				length = rbBiggestEntry * IndexEntryTime.ENTRY_LEN + IndexEntryTime.ENTRY_LEN;
				System.arraycopy(oldBuffer, 0, rbBuffer, start, length);
				if(rbFirstUnsaved != NO_VALUE) rbFirstUnsaved = rbFirstUnsaved - ((oldBuffer.length / IndexEntryTime.ENTRY_LEN) - newSize);
				rbBiggestEntry = (start + length) / IndexEntryTime.ENTRY_LEN - 1;
				rbSmallestEntry = 0;
			}
		}
	}


	private int rbElementsInbetween(int start, int stop, int ringbufferSize) {
		if(start <= stop) return stop - start;
		return (ringbufferSize + (stop - start)) + 1;
	}


	private int rbEntries() {
		if(rbBiggestEntry == -1) {
			return 0;
		}
		else if(rbSmallestEntry > rbBiggestEntry) {
			return rbCapacity() - rbSmallestEntry + rbBiggestEntry + 1;
		}
		else {
			return rbBiggestEntry - rbSmallestEntry + 1;
		}
	}


	private int rbBytePos(int pos) {
		if(rbSmallestEntry + pos < rbCapacity()) {
			return (rbSmallestEntry + pos) * IndexEntryTime.ENTRY_LEN;
		}
		else {
			return (pos - rbCapacity() + rbSmallestEntry) * IndexEntryTime.ENTRY_LEN;
		}
	}


	private int rbCapacity() {
		return rbBuffer.length / IndexEntryTime.ENTRY_LEN;
	}


	private int rbCapacityInBytes() {
		return rbBuffer.length;
	}


	private int rbEntryLength() {
		return IndexEntryTime.ENTRY_LEN;
	}


	private boolean rbMustBeSaved() {
		return (rbFirstUnsaved == 0 && rbEntries() == rbCapacity());
	}


	private void rbAddCurrentEntry() {
		if(currentEntry[DT_MIN_POS] >= rbGetDTMin(rbEntries() - 1)) {
			rbAppendCurrentEntry();
		}
		else {
			rbInsertElement();
		}
	}


	private void rbAppendCurrentEntry() {
		if(rbEntries() == rbCapacity() && IndexEntryTime.getDTMax(rbBuffer, rbStep(rbBiggestEntry)) > biggestDTMaxInFile) {
			biggestDTMaxInFile = IndexEntryTime.getDTMax(rbBuffer, rbStep(rbBiggestEntry));
		}

		IndexEntryTime.setEntry(rbBuffer, rbStep(rbBiggestEntry), currentEntry);
		if(rbEntries() == rbCapacity() && rbStep(rbBiggestEntry) == rbSmallestEntry) { // XXX 2006-04-12 Zweite Bedingung überflüssig?
			rbSmallestEntry = rbStep(rbSmallestEntry);
			if(rbFirstUnsaved >= 0) rbFirstUnsaved = rbStepBack(rbFirstUnsaved);
		}

		rbBiggestEntry = rbStep(rbBiggestEntry);
		if(rbFirstUnsaved < 0) rbFirstUnsaved = rbEntries() - 1;
	}


	private void rbInsertElement() {
		if(rbCapacity() == rbEntries()) {
			if(biggestDTMaxInFile != NO_VALUE && IndexEntryTime.getDTMax(rbBuffer, rbSmallestEntry) > biggestDTMaxInFile) {
				biggestDTMaxInFile = IndexEntryTime.getDTMax(rbBuffer, rbSmallestEntry);
			}
		}
		int position = rbFindDTMin(currentEntry[DT_MIN_POS]);
		for(int i = rbEntries() - 1; i >= position; i--) {
			rbMoveEntryRight(i);
		}
		IndexEntryTime.setEntry(rbBuffer, (rbSmallestEntry + position) % rbCapacity(), currentEntry);

		if(rbEntries() == rbCapacity() && rbStep(rbBiggestEntry) == rbSmallestEntry) {
			rbSmallestEntry = rbStep(rbSmallestEntry);
			position = rbStepBack(position);
		}
		rbBiggestEntry = rbStep(rbBiggestEntry);

		if(rbFirstUnsaved == NO_VALUE || (rbFirstUnsaved != NO_VALUE && position < rbFirstUnsaved)) {
			rbFirstUnsaved = position;
			rbFileupdateNecessary = true;
		}
		else if(rbEntries() == rbCapacity() && rbFirstUnsaved > 0) rbFirstUnsaved = rbStepBack(rbFirstUnsaved);
	}


	private void rbExtractMaxima() {
//        long tmpEntry[] = new long[] {NO_VALUE,NO_VALUE,NO_VALUE,NO_VALUE,NO_VALUE};
		long tmpEntry[] = new long[currentEntry.length];
		System.arraycopy(currentEntry, 0, tmpEntry, 0, tmpEntry.length);

		int tmpIndex = -1;
		for(int i = rbEntries() - 1; i >= 0; i--) {
			if(rbGetCID(i) >= tmpEntry[CONTID_POS]) {
				tmpEntry = rbGetEntry(i);
				tmpIndex = i;
			}
		}
		currentEntry = tmpEntry;

		for(int i = tmpIndex + 1; i < rbCapacity(); i++) {
			rbMoveEntryLeft(i);
		}

		if(rbEntries() == 1) {
			rbBiggestEntry = -1; // also leerer ringbuffer
		}
		else {
			rbBiggestEntry = rbStepBack(rbBiggestEntry);
		}
	}


	private void rbMoveEntryRight(int index) {
		System.arraycopy(rbBuffer, rbBytePos(index), rbBuffer, rbBytePos(rbStep(index)), IndexEntryTime.ENTRY_LEN);
	}


	private void rbMoveEntryLeft(int index) {
		System.arraycopy(rbBuffer, rbBytePos(index), rbBuffer, rbBytePos(rbStepBack(index)), IndexEntryTime.ENTRY_LEN);
	}


	private int rbFindDTMin(long key) {
		int low = 0, high = rbEntries() - 1, mid;
		int max = high;
		while(low <= high) {
			mid = (low + high) / 2;
			long midVal = rbGetDTMin(mid);
			if(midVal > key) {
				high = mid - 1;
			}
			else if(midVal < key) {
				low = mid + 1;
			}
			else {
				while(midVal == key){
					mid++;
					if(mid>max) break;
					midVal = rbGetDTMin(mid);
				}
				return mid;
			}
		}
		return low;
	}


	private long rbGetDTMin(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryTime.getDTMin(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetDTMax(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryTime.getDTMax(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetDIMin(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryTime.getDIMin(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetDIMax(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryTime.getDIMax(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}


	private long rbGetCID(int index) {
		if(index < 0) {
			return -1;
		}
		else {
			return IndexEntryTime.getCID(rbBuffer, (rbSmallestEntry + index) % rbCapacity());
		}
	}

	private long[] rbGetEntry(int index) {
		return new long[]{rbGetDTMin(index), rbGetDTMax(index), rbGetDIMin(index), rbGetDIMax(index), rbGetCID(index)};
	}


	private int rbStep(int index) {
		return (index + 1) % rbCapacity();
	}


	private int rbStepBack(int index) {
		return (index + rbCapacity() - 1) % rbCapacity();
	}


	private void rbAppendToFile(RandomAccessFile raf) throws Exception {
		int first = (rbSmallestEntry + rbFirstUnsaved) % rbEntries();
		int last = rbBiggestEntry;
		raf.seek(raf.length());
		if(first > last) {
			raf.write(rbBuffer, first * rbEntryLength(), rbBuffer.length - (first * rbEntryLength()));
			raf.write(rbBuffer, 0, last * rbEntryLength() + rbEntryLength());
		}
		else {
			raf.write(rbBuffer, first * rbEntryLength(), (last - first) * rbEntryLength() + rbEntryLength());
		}
	}


	private long[] rbGetFileAnchor() {
		if(rbFirstUnsaved >= 0) {
			return IndexEntryTime.getEntry(rbBuffer, (rbSmallestEntry + rbFirstUnsaved) % rbCapacity());
		}
		else {
			return new long[]{-1, -1, -1, -1, -1};
		}
	}


	private boolean rbHasUnsavedData() {
		return rbFirstUnsaved >= 0;
	}


	private boolean rbCacheChanged() {
		return rbFileupdateNecessary;
	}


	private int rbGetFirstUnsaved() {
		return rbFirstUnsaved;
	}


	private String rbToString() {
		StringBuilder sb = new StringBuilder();
		sb.append("{");
		for(int i = 0; i < rbCapacity(); i++) {
			if(rbSmallestEntry == i) sb.append("[");
			sb.append(IndexEntryTime.getDTMin(rbBuffer, i) + "," + IndexEntryTime.getDTMax(rbBuffer, i) + ",");
			sb.append(IndexEntryTime.getDIMin(rbBuffer, i) + "," + IndexEntryTime.getDIMax(rbBuffer, i) + ",");
			sb.append(IndexEntryTime.getCID(rbBuffer, i));
			if(rbBiggestEntry == i) sb.append("]");
			sb.append("; ");
		}
		sb.delete(sb.length() - 2, sb.length());
		sb.append("}");
		return sb.toString();
	}
}
