/*
 *
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.index;

import de.bsvrz.sys.funclib.losb.util.ByteIO;

/**
 * Repräsentiert einen Eintrag im monotonen Index. Dieser besteht aus Datenindex-Minimum, Datenindex-Maximum, Archivzeit-Minimum, Archivzeit-Maximum,
 * Datenzeit-Minimum, Datenzeit-Maximum und der Container-ID, auf die sich die Min/Max-Werte beziehen.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Philippe Schrettenbrunner
 * @author Alexander Schmidt
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class IndexEntryMonotone {

	private static final int DI_MIN_POS = 0;

	private static final int DI_MAX_POS = ByteIO.LONG8B_LEN;

	private static final int AT_MIN_POS = 2 * ByteIO.LONG8B_LEN;

	private static final int AT_MAX_POS = 2 * ByteIO.LONG8B_LEN + ByteIO.LONG6B_LEN;

	private static final int DT_MIN_POS = 2 * ByteIO.LONG8B_LEN + 2 * ByteIO.LONG6B_LEN;

	private static final int DT_MAX_POS = 2 * ByteIO.LONG8B_LEN + 3 * ByteIO.LONG6B_LEN;

	private static final int CID_POS = 2 * ByteIO.LONG8B_LEN + 4 * ByteIO.LONG6B_LEN;

	public static final int ENTRY_LEN = CID_POS + ByteIO.LONG5B_LEN;

	/**
	 * Setzt einen Eintrag entsprechend der uebergebenen Werte an der angegebenen Stelle im angegebenen Feld.
	 *
	 * @param buf   Speicherbereich des Index
	 * @param pos   Position des Neueintrags in Bytes
	 * @param dIMin Datenindex-Minimum
	 * @param dIMax Datenindex-Maximum
	 * @param aTMin Archivzeit-Minimum
	 * @param aTMax Archivzeit-Maximum
	 * @param dTMin Datenzeit-Minimum
	 * @param dTMax Datenzeit-Maximum
	 * @param cID   Container-ID
	 */
	public static void setEntry(byte[] buf, int pos, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) {
		int startPos = ENTRY_LEN * pos;
		ByteIO.writeSignedLong8Byte(buf, startPos + DI_MIN_POS, dIMin);
		ByteIO.writeSignedLong8Byte(buf, startPos + DI_MAX_POS, dIMax);
		ByteIO.writeUnsignedLong6Byte(buf, startPos + AT_MIN_POS, aTMin);
		ByteIO.writeUnsignedLong6Byte(buf, startPos + AT_MAX_POS, aTMax);
		ByteIO.writeUnsignedLong6Byte(buf, startPos + DT_MIN_POS, dTMin);
		ByteIO.writeUnsignedLong6Byte(buf, startPos + DT_MAX_POS, dTMax);
		ByteIO.writeUnsignedLong5Byte(buf, startPos + CID_POS, cID);
	}

	/**
	 * Liefert das Datenindex-Minimum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Datenindex-Minimum
	 */
	public static long getDIdxMin(byte[] buf, int pos) {
		return ByteIO.readSignedLong8Byte(buf, ENTRY_LEN * pos);
	}

	/**
	 * Liefert das Datenindex-Maximum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Datenindex-Maximum
	 */
	public static long getDIdxMax(byte[] buf, int pos) {
		return ByteIO.readSignedLong8Byte(buf, ENTRY_LEN * pos + DI_MAX_POS);
	}

	/**
	 * Liefert das Archivzeit-Minimum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Archivzeit-Minimum
	 */
	public static long getATimeMin(byte[] buf, int pos) {
		return ByteIO.readUnsignedLong6Byte(buf, ENTRY_LEN * pos + AT_MIN_POS);
	}

	/**
	 * Liefert das Archivzeit-Maximum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Archivzeit-Maximum
	 */
	public static long getATimeMax(byte[] buf, int pos) {
		return ByteIO.readUnsignedLong6Byte(buf, ENTRY_LEN * pos + AT_MAX_POS);
	}

	/**
	 * Liefert das Archivzeit-Minimum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Archivzeit-Minimum
	 */
	public static long getDTimeMin(byte[] buf, int pos) {
		return ByteIO.readUnsignedLong6Byte(buf, ENTRY_LEN * pos + DT_MIN_POS);
	}

	/**
	 * Liefert das Archivzeit-Maximum des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Archivzeit-Maximum
	 */
	public static long getDTimeMax(byte[] buf, int pos) {
		return ByteIO.readUnsignedLong6Byte(buf, ENTRY_LEN * pos + DT_MAX_POS);
	}

	/**
	 * Liefert das Maximum des spezifizierten Indexart des Eintrags an der angegebenen Stelle.
	 *
	 * @param entrVal Indexart (Datenindex oder Archivzeit)
	 * @param buf     Speicherbereich des Index
	 * @param pos     Position des Eintrags in Eintraegen gezaehlt
	 *
	 * @return Maximum der Indexart
	 */
	public static long get(ATimeMonotoneIndex.Key entrVal, boolean min, byte[] buf, int pos) {
		switch(entrVal) {
			case DI:
				return min ? getDIdxMin(buf, pos) : getDIdxMax(buf, pos);
			case AT:
				return min ? getATimeMin(buf, pos) : getATimeMax(buf, pos);
			case DT:
				return min ? getDTimeMin(buf, pos) : getDTimeMax(buf, pos);
			default:
				return -1;
		}
	}

	/**
	 * Liefert die Container-ID des Eintrags an der angegebenen Stelle.
	 *
	 * @param buf Speicherbereich des Index
	 * @param pos Position des Eintrags in Bytes
	 *
	 * @return Container-ID des Eintrags
	 */
	public static long getCID(byte[] buf, int pos) {
		return ByteIO.readUnsignedLong5Byte(buf, ENTRY_LEN * pos + CID_POS);
	}
}
