/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.persistence.ContainerDataResult;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;

import java.util.*;

/**
 * Diese Klasse kombiniert mehrere {@link DataIterator Iteratoren} über verschiedene {@link de.bsvrz.dav.daf.main.archive.ArchiveDataKind Archivdatenarten}
 * um daraus einen einzelnen Iterator zu erstellen, der alle Daten enthält.
 *
 * Die Daten werden dabei entweder Anhand ihrer Datenzeit oder ihres Datenindex einsortiert.
 *
 * @author Kappich Systemberatung
 */
public class CombineAdkDataIterator implements DataIterator {
	private DataIterator _currentDelegate;
	private final List<DataIterator> _iterators;
	private final ArchiveOrder _order;

	/**
	 * Konstruktor
	 * @param containerSequences Die Sequenzen der verschiedenen Datenarten
	 * @throws PersistenceException
	 */
	public CombineAdkDataIterator(final Collection<DataSequence> containerSequences, ArchiveOrder order) throws PersistenceException {
		_order = order;
		final List<DataIterator> iterators = new ArrayList<>(containerSequences.size());
		for(DataSequence sequence : containerSequences) {
			DataIterator iterator = sequence.iterator();
			iterators.add(iterator);
		}
		_iterators = iterators;
		updateDelegate();
	}

	/**
	 * Konstruktor
	 * @param iterators Die Iteratoren der verschiedenen Datenarten
	 * @throws PersistenceException
	 */
	public CombineAdkDataIterator(final List<DataIterator> iterators, ArchiveOrder order) {
		_iterators = iterators;
		_order = order;
		updateDelegate();
	}

	@Override
	public void peek(final ContainerDataResult result) throws PersistenceException {
		if(_currentDelegate == null) throw new NoSuchElementException();
		_currentDelegate.peek(result);
	}

	@Override
	public long peekDataIndex() {
		if(_currentDelegate == null) throw new NoSuchElementException();
		return _currentDelegate.peekDataIndex();
	}

	@Override
	public long peekDataTime() {
		if(_currentDelegate == null) throw new NoSuchElementException();
		return _currentDelegate.peekDataTime();
	}

	@Override
	public long peekArchiveTime() {
		if(_currentDelegate == null) throw new NoSuchElementException();
		return _currentDelegate.peekArchiveTime();
	}

	@Override
	public void remove() throws PersistenceException {
		if(_currentDelegate == null) throw new NoSuchElementException();
		long lastDataIndex = _currentDelegate.peekDataIndex();
		_currentDelegate.remove();
		updateDelegate();
		while(!isEmpty() && _currentDelegate.peekDataIndex() == lastDataIndex) {
			_currentDelegate.remove();
			updateDelegate();
		}
	}

	private void updateDelegate() {
		if(_order == ArchiveOrder.BY_INDEX) {
			_currentDelegate = _iterators.stream().filter(it -> !it.isEmpty()).min(Comparator.comparing(DataIterator::peekDataIndex)).orElse(null);
		}
		else {
			_currentDelegate = _iterators.stream().filter(it -> !it.isEmpty()).min(Comparator.comparing(DataIterator::peekDataTime).thenComparing(DataIterator::peekDataIndex)).orElse(null);
		}
	}

	@Override
	public boolean isEmpty() {
		return _currentDelegate == null;
	}

	@Override
	public void close() throws PersistenceException {
		for(DataIterator iterator : _iterators) {
			iterator.close();
		}
	}

	@Override
	public long getCurContFileIdx() {
		if(_currentDelegate == null) throw new NoSuchElementException();
		return _currentDelegate.getCurContFileIdx();
	}
}
