/*
 *
 * Copyright 2017 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars.  If not, see <http://www.gnu.org/licenses/>.

 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436 
 * mail: <info@kappich.de>
 */

package de.bsvrz.ars.ars.persistence.iter;

import de.bsvrz.ars.ars.mgmt.ArchiveManager;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.IdDataIdentification;
import de.bsvrz.ars.ars.persistence.PersistenceException;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.index.IndexException;
import de.bsvrz.ars.ars.persistence.index.result.ContainerIndexResult;
import de.bsvrz.ars.ars.persistence.index.result.IndexResult;
import de.bsvrz.dav.daf.main.archive.*;
import de.bsvrz.sys.funclib.kappich.annotations.Nullable;

import java.io.File;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

/**
 * Klasse zum Iterieren über Datensätze einer Datenidentifikation. Diese Klasse kümmert sich darum, die benötigten Container zu suchen und bei Bedarf aus diesen die richtigen
 * Datensätze zu extrahieren.
 *
 * @author Kappich Systemberatung
 */
public class QueryDataSequence implements DataSequence {
	private final ArchiveManager _archiveManager;
	private final IdDataIdentification _dataIdentification;
	private final ArchiveDataKindCombination _archiveDataKinds;

	@Nullable
	private final ArchiveTimeSpecification _archiveTimeSpecification;
	private final DeletedTreatment _deletedTreatment;
	private final Task _task;
	private final byte[] _tempMem;
	private final PersistenceManager _persMgr;
	private final DataIdentNode _din;
	private final List<DataSequence> _containerSequences = new ArrayList<>();
	private final ArchiveOrder _order;

	/**
	 * Erstellt eine neue QueryDataSequence. Dies ist eine allgemeine Anfrageschnittstelle, über die für eine Datenidentifikation und einen Zeitbereich sowie eine Angabe der
	 * Datenarten ein {@link DataIterator Iterator} über alle in diesem Bereich vorliegenden Daten erstellt werden kann.
	 */
	public QueryDataSequence(ArchiveManager archiveManager, IdDataIdentification dataIdentification, ArchiveDataKindCombination archiveDataKinds, @Nullable ArchiveTimeSpecification archiveTimeSpecification, final ArchiveOrder order, final DeletedTreatment deletedTreatment, final Task task, final byte[] tempMem) throws Exception {
		_archiveManager = archiveManager;
		_dataIdentification = dataIdentification;
		_archiveDataKinds = archiveDataKinds;
		_archiveTimeSpecification = archiveTimeSpecification;
		_deletedTreatment = deletedTreatment;
		_task = task;
		_tempMem = tempMem;
		DataIdentTree didTree = _archiveManager.getDataIdentTree();
		_din = didTree.existsCreate(dataIdentification);
		_persMgr = _archiveManager.getPersMgr();
		boolean criticalAreaEntered = false;
		try {
			if(_din.enterCriticalArea()) {
				criticalAreaEntered = true;

				if(archiveDataKinds.isOnline()) {
					DataSequence containerSequence = getContainerSequence(ArchiveDataKind.ONLINE);
					if(containerSequence != null) {
						_containerSequences.add(containerSequence);
					}
				}
				if(archiveDataKinds.isOnlineDelayed()) {
					DataSequence containerSequence = getContainerSequence(ArchiveDataKind.ONLINE_DELAYED);
					if(containerSequence != null) {
						_containerSequences.add(containerSequence);
					}
				}
				if(archiveDataKinds.isRequested()) {
					DataSequence containerSequence = getContainerSequence(ArchiveDataKind.REQUESTED);
					if(containerSequence != null) {
						_containerSequences.add(containerSequence);
					}
				}
				if(archiveDataKinds.isRequestedDelayed()) {
					DataSequence containerSequence = getContainerSequence(ArchiveDataKind.REQUESTED_DELAYED);
					if(containerSequence != null) {
						_containerSequences.add(containerSequence);
					}
				}
			}
		}
		finally {
			if(criticalAreaEntered) {
				_din.exitCriticalArea();
			}
		}

		_order = order;
	}

	@Nullable
	private DataSequence getContainerSequence(ArchiveDataKind adk)
			throws Exception {
		if(!_din.indexesAssigned(adk)) _persMgr.assignIndexes(_archiveManager.getDataIdentTree(), _dataIdentification, adk, _task, _tempMem);

		String pathName = _persMgr.getPathName(_dataIdentification, adk);
		File adkPath = new File(pathName);

		ArchiveTimeSpecification ats = _archiveTimeSpecification;
		if(ats == null) {
			ats = new ArchiveTimeSpecification(TimingType.DATA_INDEX, false, 0, Long.MAX_VALUE);
		}
		else if(ats.isStartRelative()) {
			ats = createAbsoluteQuery(ats);
		}
		IndexResult<?> indexResult = getIndexResult(adk, ats, pathName);
		if(indexResult.isEmpty()) return null;
		ArrayList<IndexedContainer> indexedContainers = new ArrayList<>(indexResult.size());
		for(int i = 0; i < indexResult.size(); i++) {
			ContainerIndexResult result = indexResult.get(i);
			File contFile = new File(adkPath, PersistenceManager.getContFileName(result.getContainerId()));
			indexedContainers.add(new IndexedContainer(result.getContainerId(), result.getDataIndexMinimum(), result.getDataIndexMaximum(), contFile));
		}
		if(adk.isRequested()){
			// Die Container sind nicht zwingenderweise nach Datenindex sortiert, es kann Überschneidungen geben
			// Container, sie sich vom Index her überschneiden müssen daher zusammen geöffnet und betrachtet werden
			// dazu werden sie in Buckets einsortiert, die nacheinander abgearbeitet werden
			return new BucketContainerSequence(indexedContainers, adk);
		}
		else {
			indexedContainers.sort(Comparator.comparing(IndexedContainer::getContainerId));
			// Einfacher Fall: Die Container sind monoton nach Datenindex sortiert
			return new SimpleContainerSequence(indexedContainers, adk);
		}
	}

	private ArchiveTimeSpecification createAbsoluteQuery(final ArchiveTimeSpecification ats) {
		throw new UnsupportedOperationException("Derzeit keine Relativanfragen möglich");
	}

	private IndexResult<?> getIndexResult(final ArchiveDataKind adk, final ArchiveTimeSpecification ats, final String pathName) throws IndexException {
		final DataIdentNode din = _din;
		if(ats.getTimingType().equals(TimingType.DATA_INDEX)) {
			return din.getDataIdxContIDs(adk, pathName, _tempMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else if(ats.getTimingType().equals(TimingType.DATA_TIME)) {
			return din.getDataTimeContIDs(adk, pathName, _tempMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else if(ats.getTimingType().equals(TimingType.ARCHIVE_TIME)) {
			return din.getArSTimeContIDs(adk, pathName, _tempMem, ats.getIntervalStart(), ats.getIntervalEnd());
		}
		else {
			return IndexResult.EMPTY;
		}
	}

	@Override
	public DataIterator iterator() throws PersistenceException {
		CombineAdkDataIterator iterator = new CombineAdkDataIterator(_containerSequences, _order);
		if(_archiveTimeSpecification == null) return iterator;

		// wenn ein Zeitbereich angegeben wurde, das Ergebnis so Filtern, dass es auf diesen Zeitbereich beschränkt ist.
		return new TimeSpecificationFilterIterator(iterator, _archiveTimeSpecification);
	}

	private class SimpleContainerSequence implements DataSequence {
		private final List<IndexedContainer> _containerFiles;
		private final ArchiveDataKind _adk;

		public SimpleContainerSequence(final List<IndexedContainer> containerFiles, final ArchiveDataKind adk) {
			_containerFiles = containerFiles;
			_adk = adk;
		}

		@Override
		public DataIterator iterator() throws PersistenceException {
			return new SequentialContainerIterator(_archiveManager, _containerFiles, _dataIdentification, _adk, _deletedTreatment, _task);
		}
	}

	private class BucketContainerSequence implements DataSequence {
		// Gruppen von sich zeitlich überschneidenden Containern aufsteigend nach Datenindex sortiert
		private final List<List<IndexedContainer>> _containerFiles;
		private final ArchiveDataKind _adk;

		public BucketContainerSequence(final List<IndexedContainer> containerFiles, final ArchiveDataKind adk) {
			_containerFiles = new ArrayList<>();
			ArrayList<IndexedContainer> tmpBlock = new ArrayList<>();
			ArrayList<IndexedContainer> tmp = new ArrayList<>(containerFiles);
			tmp.sort(Comparator.comparing(IndexedContainer::getDataIndexMinimum).reversed());
			// tmp ist absteigend nach minimalem Datenindex sortiert, da vom Ende einer ArrayList entfernen performanter ist.
			while(!tmp.isEmpty()) {
				IndexedContainer last = tmp.remove(tmp.size() - 1);
				tmpBlock.add(last);
				long min = last.getDataIndexMinimum();
				long max = last.getDataIndexMaximum();
				while(true) {
					boolean doContinue = false;
					for(Iterator<IndexedContainer> iterator = tmp.iterator(); iterator.hasNext(); ) {
						final IndexedContainer indexedContainer = iterator.next();
						if(overlap(indexedContainer, min, max)) {
							iterator.remove();
							tmpBlock.add(indexedContainer);
							min = Math.min(min, indexedContainer.getDataIndexMinimum());
							max = Math.max(max, indexedContainer.getDataIndexMaximum());
							doContinue = true;
						}
					}
					if(!doContinue) break;
				}
				_containerFiles.add(new ArrayList<>(tmpBlock));
				tmpBlock.clear();
			}
			_adk = adk;
		}

		private boolean overlap(final IndexedContainer indexedContainer, final long min, final long max) {
			return indexedContainer.getDataIndexMinimum() <= max && indexedContainer.getDataIndexMaximum() >= min;
		}


		@Override
		public DataIterator iterator() throws PersistenceException {
			return new BucketContainerIterator(_archiveManager, _containerFiles, _dataIdentification, _adk, _deletedTreatment, _task, _order);
		}
	}

	public static class IndexedContainer {
		private final long _containerId;
		private final long _dataIndexMinimum;
		private final long _dataIndexMaximum;
		private final File _containerFile;

		public IndexedContainer(final long containerId, final long dataIndexMinimum, final long dataIndexMaximum, final File containerFile) {
			_containerId = containerId;
			_dataIndexMinimum = dataIndexMinimum;
			_dataIndexMaximum = dataIndexMaximum;
			_containerFile = containerFile;
		}

		public long getContainerId() {
			return _containerId;
		}

		public long getDataIndexMinimum() {
			return _dataIndexMinimum;
		}

		public long getDataIndexMaximum() {
			return _dataIndexMaximum;
		}

		public File getContainerFile() {
			return _containerFile;
		}
	}

}
